#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes

import numpy as np
import scipy.special

from ciderpress import lib
from ciderpress.dft.grids_indexer import AtomicGridsIndexer
from ciderpress.lib import load_library as load_cider_library

libcider = load_cider_library("libmcider")
libcider.get_atco_nbas.restype = ctypes.c_int

DEFAULT_AUX_BETA = 1.6
DEFAULT_CIDER_LMAX = 10
ATOM_OF = 0
ANG_OF = 1
PTR_EXP = 5
PTR_COEFF = 6
BAS_WIDTH = 8

IFEAT_ID_TO_CONTRIB = {
    0: 0,  # squared-exponential
    1: 1,  # R2 x squared-exp
    2: 2,  # AR2 x squared-exp
    3: 7,  # A x squared-exp
    4: 8,  # A2R2 x squared-exp
    5: 9,  # Laplacian
    6: (4, 5),  # DR x squared-exp
    7: (3, 6),  # A DR x squared-exp
}


def atom_loc_from_bas(bas):
    natm = np.max(bas[:, ATOM_OF]) + 1  # max ia + 1 is number of atoms
    atom_loc = np.zeros(natm + 1, dtype=np.int32)
    for b in bas:
        atom_loc[b[ATOM_OF] + 1] += 1
    atom_loc = np.cumsum(atom_loc)
    return np.asarray(atom_loc, dtype=np.int32, order="C")


# from PySCF, helper function for normalization coeff
def gaussian_int(n, alpha):
    n1 = (n + 1) * 0.5
    return scipy.special.gamma(n1) / (2.0 * alpha**n1)


# from PySCF, helper function for normalization coeff
def gto_norm(l, expnt):
    if np.all(l >= 0):
        return 1 / np.sqrt(gaussian_int(l * 2 + 2, 2 * expnt))
    else:
        raise ValueError("l should be >= 0")


def get_etb_from_expnt_range(
    lmax, beta, emin_by_l, emax_by_l, def_amax, def_amin, lower_fac=1.0, upper_fac=1.0
):
    emin_by_l = np.array(emin_by_l)
    emax_by_l = np.array(emax_by_l)
    l_max1 = np.min(emin_by_l.size)
    assert emin_by_l.size == emax_by_l.size
    # Estimate the exponents ranges by geometric average
    emax = np.sqrt(np.einsum("i,j->ij", emax_by_l, emax_by_l))
    emin = np.sqrt(np.einsum("i,j->ij", emin_by_l, emin_by_l))
    liljsum = np.arange(l_max1)[:, None] + np.arange(l_max1)
    emax_by_l = [emax[liljsum == ll].max() for ll in range(l_max1 * 2 - 1)]
    emin_by_l = [emin[liljsum == ll].min() for ll in range(l_max1 * 2 - 1)]
    # Tune emin and emax
    emin_by_l = np.array(emin_by_l) * lower_fac  # *2 for alpha+alpha on same center
    emax_by_l = np.array(emax_by_l) * upper_fac  # / (np.arange(l_max1*2-1)*.5+1)

    cond = emax_by_l == 0
    emax_by_l[cond] = def_amax
    emin_by_l[cond] = def_amin
    emax_by_l = np.maximum(def_amax, emax_by_l[: lmax + 1])
    emin_by_l = np.minimum(def_amin, emin_by_l[: lmax + 1])

    nmaxs = np.ceil(np.log(emax_by_l) / np.log(beta))
    nmins = np.floor(np.log(emin_by_l) / np.log(beta))
    emin_by_l = beta**nmins
    ns = nmaxs - nmins + 1
    etb = []
    for l, n in enumerate(np.ceil(ns).astype(int)):
        # print(l, n, emin_by_l[l], emax_by_l[l], beta)
        if n > 0 and l <= lmax:
            etb.append((l, n, emin_by_l[l], beta))
    return etb


def get_gamma_lists_from_bas_and_env(bas, env):
    natm = np.max(bas[:, ATOM_OF]) + 1
    atom2l0 = [0]
    lmaxs = []
    gamma_loc = [0]
    all_exps = []
    all_coefs = []
    for ia in range(natm):
        ls = bas[bas[:, ATOM_OF] == ia, ANG_OF]
        # ls must be ascending
        assert np.all(ls[1:] >= ls[:-1])
        lmax = np.max(ls)
        exps = []
        coefs = []
        lmaxs.append(np.max(ls))
        for l in range(lmax + 1):
            cond = np.logical_and(bas[:, ATOM_OF] == ia, bas[:, ANG_OF] == l)
            exps_l = env[bas[cond, PTR_EXP]]
            coefs_l = env[bas[cond, PTR_COEFF]]
            gamma_loc.append(len(exps_l))
            exps.append(exps_l)
            coefs.append(coefs_l)
        all_exps.append(np.concatenate(exps))
        all_coefs.append(np.concatenate(coefs))
        atom2l0.append(len(gamma_loc) - 1)
    gamma_loc = np.asarray(np.cumsum(gamma_loc), dtype=np.int32, order="C")
    lmaxs = np.asarray(lmaxs, dtype=np.int32, order="C")
    all_exps = np.asarray(np.concatenate(all_exps), dtype=np.float64, order="C")
    all_coefs = np.asarray(np.concatenate(all_coefs), dtype=np.float64, order="C")
    atom2l0 = np.asarray(atom2l0, dtype=np.int32, order="C")
    return atom2l0, lmaxs, gamma_loc, all_coefs, all_exps


def get_gamma_lists_from_etb_list(etb_list):
    natm = len(etb_list)
    atom2l0 = [0]
    lmaxs = []
    gamma_loc = [0]
    all_exps = []
    all_coefs = []
    for ia in range(natm):
        etb = etb_list[ia]
        lmax = -1
        exps = []
        coefs = []
        for l, nexp, emin, beta in etb:
            assert l == lmax + 1, "l values must be in ascending order"
            lmax = l
            exps_l = emin * beta ** (np.flip(np.arange(nexp)).astype(np.float64))
            coefs_l = gto_norm(l, exps_l)
            gamma_loc.append(len(exps_l))
            exps.append(exps_l)
            coefs.append(coefs_l)
        all_exps.append(np.concatenate(exps))
        all_coefs.append(np.concatenate(coefs))
        atom2l0.append(len(gamma_loc) - 1)
        lmaxs.append(lmax)
    gamma_loc = np.asarray(np.cumsum(gamma_loc), dtype=np.int32, order="C")
    lmaxs = np.asarray(lmaxs, dtype=np.int32, order="C")
    all_exps = np.asarray(np.concatenate(all_exps), dtype=np.float64, order="C")
    all_coefs = np.asarray(np.concatenate(all_coefs), dtype=np.float64, order="C")
    atom2l0 = np.asarray(atom2l0, dtype=np.int32, order="C")
    return atom2l0, lmaxs, gamma_loc, all_coefs, all_exps


def get_convolution_expnts_from_expnts(
    gammas, atom2l0, lmaxs, gamma_loc, all_exps, gbuf=np.inf
):
    new_gamma_loc = [0]
    new_all_exps = []
    new_all_coefs = []
    for ia in range(len(atom2l0) - 1):
        lmax = lmaxs[ia]
        new_exps = []
        new_coefs = []
        for l in range(lmax + 1):
            loc0 = gamma_loc[atom2l0[ia] + l]
            loc1 = gamma_loc[atom2l0[ia] + l + 1]
            old_exps = all_exps[loc0:loc1]
            gmax = np.max(old_exps)
            cond = gammas < gmax * gbuf
            new_exps_l = gammas[cond]
            new_coefs_l = gto_norm(l, gammas[cond])
            new_gamma_loc.append(len(new_exps_l))
            new_exps.append(new_exps_l)
            new_coefs.append(new_coefs_l)
        new_all_exps.append(np.concatenate(new_exps))
        new_all_coefs.append(np.concatenate(new_coefs))
    gamma_loc = np.asarray(np.cumsum(new_gamma_loc), dtype=np.int32, order="C")
    lmaxs = np.asarray(lmaxs, dtype=np.int32, order="C")
    all_exps = np.asarray(np.concatenate(new_all_exps), dtype=np.float64, order="C")
    all_coefs = np.asarray(np.concatenate(new_all_coefs), dtype=np.float64, order="C")
    assert all_exps.size == all_coefs.size == gamma_loc[-1]
    atom2l0 = np.asarray(atom2l0, dtype=np.int32, order="C")
    return atom2l0, lmaxs, gamma_loc, all_coefs, all_exps


def get_reciprocal_expnts_from_expnts(atom2l0, lmaxs, gamma_loc, all_coefs, all_exps):
    new_all_exps = []
    new_all_coefs = []
    for ia in range(len(atom2l0) - 1):
        lmax = lmaxs[ia]
        new_exps = []
        new_coefs = []
        for l in range(lmax + 1):
            loc0 = gamma_loc[atom2l0[ia] + l]
            loc1 = gamma_loc[atom2l0[ia] + l + 1]
            old_exps = all_exps[loc0:loc1]
            exps = 1.0 / (4 * old_exps)
            old_coefs = all_coefs[loc0:loc1]
            const = 2 ** (-2 - l) * np.sqrt(np.pi)
            coefs = old_coefs * const * old_exps ** (-1.5 - l)
            new_exps.append(exps)
            new_coefs.append(coefs)
        new_all_exps.append(np.concatenate(new_exps))
        new_all_coefs.append(np.concatenate(new_coefs))
    lmaxs = np.asarray(lmaxs, dtype=np.int32, order="C")
    all_exps = np.asarray(np.concatenate(new_all_exps), dtype=np.float64, order="C")
    all_coefs = np.asarray(np.concatenate(new_all_coefs), dtype=np.float64, order="C")
    assert all_exps.size == all_coefs.size == gamma_loc[-1]
    atom2l0 = np.asarray(atom2l0, dtype=np.int32, order="C")
    return atom2l0, lmaxs, gamma_loc, all_coefs, all_exps


class ATCBasis:
    """
    A wrapper for the atc_basis_set C object, which is used to store
    a simple (but usually large), uncontracted Gaussian basis along
    with various index lists and other data for use in performing
    projections and manipulations of data in this basis on individual
    atoms.
    """

    def __init__(self, atom2l0, lmaxs, gamma_loc, coefs, exps, lower=True):
        """
        Initialize ATCBasis.

        Args:
            atom2l0 (np.ndarray): int32 array of length natm + 1 such that
                gamma_loc[atom2l0[ia] : atom2l0[ia + 1]] contains the global
                gamma_loc indexes for atom ia.
            lmaxs (np.ndarray) : int32, maximum l value for each atom.
            gamma_loc (np.ndarray): int32, a shell indexing list.
                gamma_loc[atom2l0[ia] + l] is the index of the first shell of
                atom ia with angular momentum l.
                [gamma_loc[atom2l0[ia] + l], atom2lo[ia] + l + 1]) is the range
                of indexes with atom ia and angular momentum l. These ranges
                are used to index coefs and exps.
            coefs (np.ndarray): float64, list of coefficients for
                normalizing Gaussian basis functions.
            exps (np.ndarray): float64, list of exponents for Gaussian
                basis functions.
            lower (bool): Whether to store lower or upper matrix for
                Cholesky factorizations within the C atc_basis_set object.
        """
        atco = lib.c_null_ptr()
        gamma_loc = np.asarray(gamma_loc, dtype=np.int32, order="C")
        lmaxs = np.asarray(lmaxs, dtype=np.int32, order="C")
        exps = np.asarray(exps, dtype=np.float64, order="C")
        coefs = np.asarray(coefs, dtype=np.float64, order="C")
        atom2l0 = np.asarray(atom2l0, dtype=np.int32, order="C")
        self.natm = len(atom2l0) - 1
        libcider.generate_atc_basis_set(
            ctypes.byref(atco),
            atom2l0.ctypes.data_as(ctypes.c_void_p),
            lmaxs.ctypes.data_as(ctypes.c_void_p),
            gamma_loc.ctypes.data_as(ctypes.c_void_p),
            exps.ctypes.data_as(ctypes.c_void_p),
            coefs.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(len(atom2l0) - 1),
            ctypes.c_char(b"L" if lower else b"U"),
        )
        self._lower = lower
        self._atco = atco

    def __del__(self):
        """Free C data associated with this object."""
        libcider.free_atc_basis_set(self._atco)

    @property
    def atco_c_ptr(self):
        """Get a pointer to the C data for this object."""
        return self._atco

    @property
    def ao_loc(self):
        ao_loc = np.cumsum(2 * self.bas[:, ANG_OF] + 1)
        ao_loc = np.append([0], ao_loc)
        return np.asarray(ao_loc, dtype=np.int32, order="C")

    @property
    def atom_loc(self):
        return atom_loc_from_bas(self.bas)

    @property
    def nao(self):
        """Get the number of atomic orbitals in this atom-centered basis."""
        return libcider.get_atco_nao(self._atco)

    @property
    def nbas(self):
        """Get the number of shells in this atom-centered basis."""
        return libcider.get_atco_nbas(self._atco)

    @property
    def bas(self):
        """Get a copy of the shell information for this basis."""
        bas = np.zeros((self.nbas, 8), order="C", dtype=np.int32)
        libcider.get_atco_bas(
            bas.ctypes.data_as(ctypes.c_void_p),
            ctypes.cast(self._atco, ctypes.c_void_p),
        )
        return bas

    @property
    def env(self):
        """Get a copy of the exponents and coefficients for this basis."""
        nbas = libcider.get_atco_nbas(self._atco)
        env = np.zeros((2 * nbas,), order="C", dtype=np.float64)
        libcider.get_atco_env(env.ctypes.data_as(ctypes.c_void_p), self._atco)
        return env

    def get_reciprocal_atco(self):
        args = get_gamma_lists_from_bas_and_env(self.bas, self.env)
        args = get_reciprocal_expnts_from_expnts(*args)
        return ATCBasis(*args, lower=True)

    def convert_rad2orb_(
        self, theta_rlmq, p_uq, loc, rads, rad2orb=True, offset=None, zero_output=True
    ):
        """
        Take a set of functions stored on radial grids and spherical harmonics
        and project it onto the atomic orbital basis. This is an in-place
        operation, where theta_rlmq is written if rad2orb is False and
        p_uq is written if rad2orb is True.
        NOTE: This is an in-place operation, and it is ADDITIVE, i.e.
        the output is added to rather than overwritten and therefore
        must be initialized before passing as an argument.
        - If rad2orb is True, loc has
          shape (natm + 1,), and theta_rlmq[loc[ia] : loc[ia+1]] is
          the set of functions on atom ia. This corresponds to
          the ra_loc member of an AtomicGridsIndexer
        - If rad2orb is False, loc has shape (nrad,), and theta_rlmq[r]
          is located on atom loc[r]. This corresponds to the ar_loc
          member of an AtomicGridsIndexer

        Args:
            theta_rlmq (np.ndarray): float64, shape (nrad, nlm, nalpha)
            p_uq (np.ndarray): float64, shape (ngrids, nalpha)
            loc (np.ndarray, AtomicGridsIndexer):
                An int32 array whose shape depends on whether rad2orb is True.
            rads (np.ndarray): float64 list of radial coordinates with size
                theta_rlmq.shape[0]. Corresponds to rad_arr member of
                AtomicGridsIndexer.
            rad2orb (bool): Whether to convert radial coordinates to orbital
                basis (True) or vice versa (False).
            offset (int): starting index in p_uq to add to/from
            zero_output (bool): Whether to set output zero before
                adding to it. Default to true. Set to False if you
                want to add to the existing values in the output
                array.
        """
        if isinstance(loc, AtomicGridsIndexer):
            if rad2orb:
                loc = loc.ra_loc
            else:
                loc = loc.ar_loc
        nalpha = theta_rlmq.shape[-1]
        stride = p_uq.shape[1]
        if offset is None:
            offset = 0
        else:
            assert offset >= 0
        nrad = rads.size
        assert loc.flags.c_contiguous
        assert loc.dtype == np.int32
        assert loc.ndim == 1
        assert rads.flags.c_contiguous
        assert rads.dtype == np.float64
        assert theta_rlmq.ndim == 3
        assert theta_rlmq.shape[0] == nrad
        assert theta_rlmq.flags.c_contiguous
        assert theta_rlmq.dtype == np.float64
        assert p_uq.ndim == 2
        assert p_uq.shape[0] == self.nao
        assert p_uq.flags.c_contiguous
        assert p_uq.dtype == np.float64
        assert offset + nalpha <= stride
        if rad2orb:
            assert loc.size == self.natm + 1
            fn = libcider.contract_rad_to_orb
            if zero_output:
                p_uq[:, offset : offset + nalpha] = 0.0
        else:
            assert loc.size == nrad
            fn = libcider.contract_orb_to_rad
            if zero_output:
                theta_rlmq[:] = 0.0
        fn(
            theta_rlmq.ctypes.data_as(ctypes.c_void_p),
            p_uq.ctypes.data_as(ctypes.c_void_p),
            loc.ctypes.data_as(ctypes.c_void_p),
            rads.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(rads.size),
            ctypes.c_int(theta_rlmq.shape[1]),
            self.atco_c_ptr,
            ctypes.c_int(nalpha),
            ctypes.c_int(stride),
            ctypes.c_int(offset),
        )


class ConvolutionCollection:
    """
    ConvolutionCollection is an object to assist in performing
    nonlocal density feature convolutions in a Gaussian basis set.
    It stores Gaussian integrals corresponding to convolutions
    of the input basis atco_inp by the squared-exponential
    functions with exponents alphas, projected onto the output
    basis atco_out. One can perform forward and backward convolutions.
    """

    is_vk = False

    def __init__(
        self,
        atco_inp,
        atco_out,
        alphas,
        alpha_norms,
        has_vj=True,
        ifeat_ids=None,
    ):
        """
        Initialize ConvolutionCollection.

        Args:
            atco_inp (ATCBasis): Input basis for convolutions
            atco_out (ATCBasis): Output basis for convolutions
            alphas (np.ndarray): float64 list of interpolating exponents
            alpha_norms (np.ndarray): float64 list of normalization
                factors for interpolation exponent functions.
            has_vj (bool): Whether to perform version j convolutions.
            ifeat_ids (list): Version i feature contributions to compute.
                Must be in strictly increasing order so that output
                can be processed by interpolation routines.
        """
        self.atco_inp = atco_inp
        self.atco_out = atco_out
        self._alphas = np.ascontiguousarray(alphas, dtype=np.float64)
        self._alpha_norms = np.ascontiguousarray(alpha_norms, dtype=np.float64)
        self._nalpha = self._alphas.size
        assert self._alpha_norms.size == self._nalpha
        if ifeat_ids is None:
            ifeat_ids = []
        icontrib0_ids = []
        icontrib1m_ids = []
        icontrib1p_ids = []
        for ifid in ifeat_ids:
            if ifid not in IFEAT_ID_TO_CONTRIB:
                raise ValueError("Unrecognized feature")
            icid = IFEAT_ID_TO_CONTRIB[ifid]
            if isinstance(icid, int):
                if len(icontrib1m_ids) > 0:
                    raise ValueError("l0 features must preceed l1 in ifeat_ids")
                icontrib0_ids.append(icid)
            else:
                assert len(icid) == 2
                icontrib1m_ids.append(icid[0])
                icontrib1p_ids.append(icid[1])
        self._ifeat_ids = ifeat_ids
        self._icontrib0_ids = icontrib0_ids
        self._icontrib1m_ids = icontrib1m_ids
        self._icontrib1p_ids = icontrib1p_ids
        self._icontrib_ids = np.ascontiguousarray(
            np.concatenate([icontrib0_ids, icontrib1m_ids, icontrib1p_ids]),
            dtype=np.int32,
        )
        if len(self._icontrib_ids) == 0 and not has_vj:
            raise ValueError("Trying to initialize an empty collection")
        self._nbeta = len(self._icontrib_ids)
        if has_vj:
            self._nbeta += self._nalpha
        self._has_vj = has_vj
        self._integrals_initialized = False
        self._ccl = lib.c_null_ptr()
        libcider.generate_convolution_collection(
            ctypes.byref(self._ccl),
            self.atco_inp.atco_c_ptr,
            self.atco_out.atco_c_ptr,
            self._alphas.ctypes.data_as(ctypes.c_void_p),
            self._alpha_norms.ctypes.data_as(ctypes.c_void_p),
            self._icontrib_ids.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(self._nalpha),
            ctypes.c_int(len(self._icontrib_ids)),
            ctypes.c_int(1 if self._has_vj else 0),
        )

    def __del__(self):
        """Free C data associated with this object."""
        libcider.free_convolution_collection(self._ccl)

    @property
    def n0(self):
        n0 = len(self._icontrib0_ids)
        if self._has_vj:
            n0 += self.nalpha
        return n0

    @property
    def n1(self):
        return len(self._icontrib1m_ids)

    @property
    def nalpha(self):
        """Number of inputs for the convolutions."""
        return self._nalpha

    @property
    def nbeta(self):
        """Total number of outputs of the convolutions."""
        return self._nbeta

    @property
    def num_out(self):
        return self.nbeta

    def compute_integrals_(self):
        """Compute and store the Gaussian integrals for
        performing convolutions."""
        if not self._integrals_initialized:
            libcider.generate_atc_integrals_all(self._ccl)
            self._integrals_initialized = True

    def solve_projection_coefficients(self):
        """Multiply convolution integral tensor by the inverse
        overlap matrices of the input and output bases."""
        if not self._integrals_initialized:
            self.compute_integrals_()
        libcider.solve_atc_coefs(self._ccl)

    def multiply_atc_integrals(self, input, output=None, fwd=True):
        """
        Perform the convolutions from the input orbital basis
        to the output orbital basis. If fwd is True, the input orbital
        basis is atco_inp and the output orbital basis is atco_out.
        If fwd is False, the input orbital basis is atco_out and the
        output orbital basis is atco_inp (i.e. the convolution is done
        in reverse.)

        Args:
            input (np.ndarray) : float64 array with shape
                (atco_inp, nalpha) for fwd=True and (atco_out, nbeta)
                for fwd=False
            output (np.ndarray, optional) : float64 array with shape
                (atco_out, nbeta) for fwd=True and (atco_inp, nalpha)
                for fwd=False. NOTE: If passed, should be all zeros.
                If None, output is initialized within the function
                and then returned.
            fwd (bool): Whether convolution is performed forward or backward.

        Returns:
            output
        """
        if output is None:
            if fwd:
                output = np.zeros((self.atco_out.nao, self.nbeta))
            else:
                output = np.zeros((self.atco_inp.nao, self.nalpha))
        if not self._integrals_initialized:
            raise RuntimeError("Must initialize integrals before calling")
        assert input.flags.c_contiguous
        assert output.flags.c_contiguous
        if fwd:
            assert input.shape == (self.atco_inp.nao, self.nalpha), (
                input.shape,
                (self.atco_inp.nao, self.nalpha),
            )
            assert output.shape == (self.atco_out.nao, self.nbeta), (
                output.shape,
                (self.atco_out.nao, self.nbeta),
            )
        else:
            assert output.shape == (self.atco_inp.nao, self.nalpha), (
                output.shape,
                (self.atco_inp.nao, self.nalpha),
            )
            assert input.shape == (self.atco_out.nao, self.nbeta), (
                input.shape,
                (self.atco_out.nao, self.nbeta),
            )
        libcider.multiply_atc_integrals(
            input.ctypes.data_as(ctypes.c_void_p),
            output.ctypes.data_as(ctypes.c_void_p),
            self._ccl,
            ctypes.c_int(1 if fwd else 0),
        )
        return output


class ConvolutionCollectionK(ConvolutionCollection):
    """
    Modified ConvolutionCollection for version k features. This is needed
    because multiplication by the ovlp_mats tensor in C is different
    for version k.
    """

    is_vk = True

    def __init__(self, atco_inp, atco_out, alphas, alpha_norms):
        """Simplified initializer for version K features.
        See ConvolutionCollection.__init__ for details. feat_id is the
        same as ifeat_ids except only one input is allowed instead of a list."""
        super(ConvolutionCollectionK, self).__init__(
            atco_inp, atco_out, alphas, alpha_norms, has_vj=False, ifeat_ids=[0]
        )

    @property
    def n0(self):
        return self.nalpha

    @property
    def num_out(self):
        return self.nalpha

    def multiply_atc_integrals(self, input, output=None, fwd=True):
        """
        Modified multiply_atc_integrals function for version k

        Args:
            input (np.ndarray) : float64 array with shape
                (atco_inp, nalpha) for fwd=True and (atco_out, nalpha)
                for fwd=False
            output (np.ndarray, optional) : float64 array with shape
                (atco_out, nalpha) for fwd=True and (atco_inp, nalpha)
                for fwd=False. NOTE: If passed, should be all zeros.
                If None, output is initialized within the function
                and then returned.
            fwd (bool): Whether convolution is performed forward or backward.

        Returns:
            output
        """
        if fwd:
            atco_inp = self.atco_inp
            atco_out = self.atco_out
        else:
            atco_inp = self.atco_out
            atco_out = self.atco_inp
        if output is None:
            output = np.zeros((atco_inp.nao, self.nalpha))
        if not self._integrals_initialized:
            raise RuntimeError("Must initialize integrals before calling")
        assert input.flags.c_contiguous
        assert output.flags.c_contiguous
        assert input.shape == (atco_inp.nao, self.nalpha)
        assert output.shape == (atco_out.nao, self.nalpha)
        libcider.multiply_atc_integrals_vk(
            input.ctypes.data_as(ctypes.c_void_p),
            output.ctypes.data_as(ctypes.c_void_p),
            self._ccl,
            ctypes.c_int(1 if fwd else 0),
        )
        return output



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import numpy as np
from ase.dft.bandgap import bandgap
from ase.units import Bohr, Ha
from gpaw.occupations import FixedOccupationNumbers, OccupationNumberCalculator
from gpaw.pw.density import ReciprocalSpaceDensity
from gpaw.sphere.lebedev import Y_nL, weight_n

from ciderpress.dft.plans import SemilocalPlan
from ciderpress.dft.settings import (
    FeatureSettings,
    FracLaplSettings,
    NLDFSettings,
    SDMXBaseSettings,
    SemilocalSettings,
)
from ciderpress.gpaw.atom_descriptor_utils import (
    PASDWCiderFeatureKernel,
    calculate_paw_sl_features,
    calculate_paw_sl_features_deriv,
)
from ciderpress.gpaw.calculator import CiderGGAHybridKernel, CiderMGGAHybridKernel
from ciderpress.gpaw.cider_paw import CiderGGA, CiderGGAPASDW, CiderMGGA, CiderMGGAPASDW


class RhoVectorSettings(SemilocalSettings):
    def __init__(self):
        self.level = "MGGA"
        self.mode = "l"

    @property
    def nfeat(self):
        return 5

    @property
    def get_feat_usps(self):
        return [3, 4, 4, 4, 5]

    def ueg_vector(self, rho=1):
        tau0 = 0.3 * (3 * np.pi**2) ** (2.0 / 3) * rho ** (5.0 / 3)
        return np.array([rho, 0, 0, 0, tau0])


def get_descriptors(
    calc,
    settings,
    p_i=None,
    use_paw=True,
    screen_dens=True,
    **kwargs,
):
    """
    Compute grid weights, feature vectors, and (optionally)
    feature vector derivatives with respect to orbital
    occupation.

    Args:
        calc: a converged GPAW calculation
        settings (BaseSettings): A settings object specifying
            which features to compute.
        p_i (optional): A list of (s, k, n) indexes for orbitals
            for which to differentiate features with respect to
            occupation
        use_paw (bool, True): Whether to use PAW corrections.
        screen_dens (bool, True): Whether to remove low-density
            grids from the return feature vectors.

    Returns:
        if p_i is None:
            weights, features
        else:
            weights, features, [list of feature derivatives]
    """
    if (
        hasattr(calc.hamiltonian.xc, "setups")
        and calc.hamiltonian.xc.setups is not None
    ):
        reset_setups = True
        contribs = []
        xccs = []
        for setup in calc.hamiltonian.xc.setups:
            xccs.append(setup.xc_correction)
            if hasattr(setup, "cider_contribs"):
                contribs.append(setup.cider_contribs)
            else:
                contribs.append(None)
    else:
        reset_setups = False
    if isinstance(settings, str):
        if settings != "l":
            raise ValueError("Settings must be Settings object or letter l")
        kcls = CiderMGGAHybridKernel
        cls = FeatSLMGGAPAW if use_paw else FeatSLMGGA
        cider_kernel = kcls(XCShell(RhoVectorSettings()), 0, "GGA_X_PBE", "GGA_C_PBE")
    elif isinstance(settings, SemilocalSettings):
        if settings.level == "MGGA":
            kcls = CiderMGGAHybridKernel
            cls = FeatSLMGGAPAW if use_paw else FeatSLMGGA
        else:
            kcls = CiderGGAHybridKernel
            cls = FeatSLGGAPAW if use_paw else FeatSLGGA
        cider_kernel = kcls(XCShell(settings), 0, "GGA_X_PBE", "GGA_C_PBE")
    elif isinstance(settings, NLDFSettings):
        if settings.sl_level == "MGGA":
            kcls = CiderMGGAHybridKernel
            cls = FeatMGGAPAW if use_paw else FeatMGGA
        else:
            kcls = CiderGGAHybridKernel
            cls = FeatGGAPAW if use_paw else FeatGGA
        cider_kernel = kcls(XCShell(settings), 0, "GGA_X_PBE", "GGA_C_PBE")
    elif isinstance(settings, FracLaplSettings):
        raise NotImplementedError(
            "Fractional Laplacian-based orbital descriptors have not yet been implemented for GPAW."
        )
    elif isinstance(settings, SDMXBaseSettings):
        raise NotImplementedError(
            "SDMX descriptors have not yet been implemented for GPAW."
        )
    else:
        raise ValueError("Unsupported settings")
    # TODO clean this up a bit
    kwargs["xmix"] = 0.0
    kwargs["encut"] = kwargs.get("qmax") or kwargs.get("encut") or 300
    kwargs["lambd"] = kwargs.get("lambd") or 1.8
    empty_xc = cls(
        cider_kernel,
        pasdw_ovlp_fit=True,
        pasdw_store_funcs=False,
        **kwargs,
    )
    empty_xc.set_grid_descriptor(calc.hamiltonian.xc.gd)
    empty_xc.initialize(calc.density, calc.hamiltonian, calc.wfs)
    empty_xc.set_positions(calc.spos_ac)
    if calc.density.nt_sg is None:
        calc.density.interpolate_pseudo_density()
    if empty_xc.orbital_dependent:
        calc.converge_wave_functions()
    if p_i is None:
        res = get_features_and_weights(calc, empty_xc, screen_dens=screen_dens)
    else:
        if use_paw:
            DD_aop = get_empty_atomic_matrix(calc.density, len(p_i))
            calculate_single_orbital_atomic_density_matrix(
                calc.wfs, DD_aop, p_i, calc.density.ncomponents
            )
        else:
            DD_aop = None
        drhodf_ixR = get_drho_df(calc, p_i)
        drhodf_ixg = interpolate_drhodf(
            empty_xc.gd, empty_xc.distribute_and_interpolate, drhodf_ixR
        )
        res = get_features_and_weights_deriv(
            calc,
            empty_xc,
            p_i,
            drhodf_ixg,
            DD_aop=DD_aop,
            screen_dens=screen_dens,
        )
    if reset_setups:
        for a, setup in enumerate(calc.hamiltonian.xc.setups):
            setup.xc_correction = xccs[a]
            if contribs[a] is None:
                setup.cider_contribs = None
            else:
                setup.cider_contribs = contribs[a]
    return res


class XCShell:
    def __init__(self, settings):
        if isinstance(settings, SemilocalSettings):
            self.settings = FeatureSettings(sl_settings=settings)
        elif isinstance(settings, NLDFSettings):
            sl_settings = SemilocalSettings(
                "npa" if settings.sl_level == "MGGA" else "np"
            )
            self.settings = FeatureSettings(
                sl_settings=sl_settings,
                nldf_settings=settings,
            )
        elif isinstance(settings, FracLaplSettings):
            self.settings = FeatureSettings(
                nlof_settings=settings,
            )
        elif isinstance(settings, SDMXBaseSettings):
            self.settings = FeatureSettings(
                sdmx_settings=settings,
            )
        else:
            raise ValueError("Unsupported settings")


def calc_fixed(bd, fref_sqn, f_qn):
    if bd.nbands == fref_sqn.shape[2]:
        nspins = fref_sqn.shape[0]
        for u, f_n in enumerate(f_qn):
            s = u % len(fref_sqn)
            q = u // nspins
            bd.distribute(fref_sqn[s, q], f_n)
    else:
        raise NotImplementedError


def calculate_single_orbital_matrix(self, D_sii, kpt, a, n):
    if kpt.rho_MM is not None:
        raise NotImplementedError
    else:
        if self.collinear:
            P_i = kpt.projections[a][n]
            assert P_i.ndim == 1
            D_sii[kpt.s] += np.dot(P_i.conj(), P_i).real
        else:
            raise NotImplementedError

    if hasattr(kpt, "c_on"):
        raise NotImplementedError


def calculate_single_orbital_atomic_density_matrix(self, D_aop, p_o, nspin):
    """
    Get the atomic density matrices for the band given
    by index tuple p

    Args:
        self: GPAW Wavefunctions object
        D_asp: stores the single-orbital density matrix
        p: (s, k, n) tuple
    """
    for a, D_op in D_aop.items():
        D_op[:] = 0.0
        for o, p in enumerate(p_o):
            # Get the apprropriate k-point index on this rank
            rank, q = self.kd.who_has(p[1])
            if self.kd.comm.rank == rank:
                s = p[0]
                u = q * nspin + s
                kpt = self.kpt_u[u]
                assert kpt.s == s
                assert kpt.k == p[1]
                D_op[o] = self.get_orbital_density_matrix(a, kpt, p[2])[s]
        self.kptband_comm.sum(D_op)


def get_empty_atomic_matrix(dens, norb):
    return dens.setups.empty_atomic_matrix(norb, dens.atom_partition)


class FixedGOccupationNumbers(FixedOccupationNumbers):
    """
    Same as FixedOccupationNumbers but with a perturb_occupation_
    function for testing incremental occupation numbers.
    """

    def perturb_occupation_(self, p, delta):
        """
        Change the occupation number of a band at every k-point
        Args:
            p: (s, k, n) or (s, n). k is ignored
            delta: amount to change occupation number
        """
        if len(p) == 3:
            p = (p[0], p[2])
        elif len(p) != 2:
            raise ValueError("Length of index tuple must be 2 or 3")
        self.f_sn[p] += delta


class FixedKOccupationNumbers(OccupationNumberCalculator):
    """
    Same as FixedOccupationNumbers but the occupations are k-point specific.
    """

    extrapolate_factor = 0.0

    def __init__(self, f_skn, kpt_indices, parallel_layout=None, weight_sk=None):
        """
        Args:
            f_skn: Occupations numbers
            parallel_layout: communictors for occupations
        """
        OccupationNumberCalculator.__init__(self, parallel_layout)
        self.kpt_indices = np.array(kpt_indices)
        self.f_skn = np.array(f_skn)
        self.weight_sk = weight_sk

    def _calculate(
        self,
        nelectrons,
        eig_qn,
        weight_q,
        f_qn,
        fermi_level_guess=np.nan,
        fix_fermi_level=False,
    ):
        calc_fixed(self.bd, self.f_skn[:, self.kpt_indices, :], f_qn)

        return np.inf, 0.0

    @classmethod
    def from_calculator(cls, calc):
        """
        Initialized from an existing Calculator
        Args:
            calc: GPAW Calculator

        Returns:
            Instance of cls
        """
        calc.wfs.calculate_occupation_numbers()
        if calc.wfs.collinear and calc.wfs.nspins == 1:
            degeneracy = 2
        else:
            degeneracy = 1
        f_un = np.array([kpt.f_n for kpt in calc.wfs.kpt_u])
        f_skn = calc.wfs.kd.collect(f_un, True)
        w_u = np.array([kpt.weightk for kpt in calc.wfs.kpt_u])[:, np.newaxis]
        weight_sk = calc.wfs.kd.collect(w_u, True) * degeneracy
        f_skn /= weight_sk
        weight_sk = weight_sk[..., 0]
        return cls(
            f_skn,
            calc.wfs.kd.get_indices(),
            parallel_layout=calc.wfs.occupations.parallel_layout,
            weight_sk=weight_sk,
        )

    def perturb_occupation_(self, p, delta, div_by_wt=False):
        """
        Change the occupation number of a band at a k-point
        Args:
            p: (s, k, n) tuple
            delta: amount to change occupation number
            div_by_wt: If True, self.weight_sk must not be None.
                Divide delta by the k-point weight so that
                finite derivatives with respect to occupation
                match analytical derivatives of changing electron number.
        """
        if div_by_wt:
            if self.weight_sk is None:
                raise ValueError("weight_sk must be set if div_by_wt=True")
            delta /= self.weight_sk[p[:2]]
        self.f_skn[p] += delta

    def todict(self):
        return {"name": "fixedk", "f_skn": self.f_skn}


def run_constant_occ_calculation_(calc):
    calc.set(occupations=FixedKOccupationNumbers.from_calculator(calc))
    etot = calc.get_potential_energy()
    return etot / Ha


def perturb_occ_(calc, p, delta):
    occ = FixedKOccupationNumbers.from_calculator(calc)
    occ.perturb_occupation_(p, delta, div_by_wt=True)
    calc.set(charge=calc.density.charge - delta)
    calc.calculate(properties=["energy"])
    calc.set(occupations=occ)
    calc.calculate(properties=["energy"])


def evaluate_perturbed_energy(calc, p, delta=1e-6, store_initial=False, reset=True):
    def _get_energy():
        calc.calculate(properties=["energy"])
        return calc.get_potential_energy() / Ha

    if store_initial:
        e_init = _get_energy()
    calc.wfs.occupations.perturb_occupation_(p, delta, div_by_wt=True)
    e_final = _get_energy()
    if reset:
        calc.wfs.occupations.perturb_occupation_(p, -1 * delta, div_by_wt=True)
    if store_initial:
        return e_final, e_init
    else:
        return e_final


def get_homo_lumo_fd_(calc, delta=1e-6):
    """
    Calculate finite difference derivatives of the
    energy wrt the VBM and CBM occupation, for comparison
    to the band gap.

    Args:
        calc: GPAW Calculator object
        delta: finite difference perturbation

    Returns:
        gap, e_vbm, e_cbm dE/df_vbm, dE/df_cbm, p_vbm, p_cbm
        All units in Ha
    """
    delta = np.abs(delta)
    gap, p_vbm, p_cbm = bandgap(calc)
    e0 = run_constant_occ_calculation_(calc)
    gap = gap / Ha
    ep_vbm, e_init = evaluate_perturbed_energy(
        calc, p_vbm, delta=-1 * delta, store_initial=True
    )
    ep_cbm = evaluate_perturbed_energy(calc, p_cbm, delta=delta, store_initial=False)
    e_vbm = calc.get_eigenvalues(kpt=p_vbm[1], spin=p_vbm[0])[p_vbm[2]] / Ha
    e_cbm = calc.get_eigenvalues(kpt=p_cbm[1], spin=p_cbm[0])[p_cbm[2]] / Ha
    assert np.abs(e_init - e0) < 1e-6, e_init - e0
    dev = (e_init - ep_vbm) / delta
    dec = (ep_cbm - e_init) / delta
    return gap, e_vbm, e_cbm, dev, dec, p_vbm, p_cbm


def get_atom_wt(functional, a):
    dv_g = functional.setups[a].xc_correction.rgd.dv_g
    weight_gn = dv_g[:, None] * weight_n
    return weight_gn.ravel()


def get_atom_feat_wt(functional):
    ae_feat, ps_feat = functional.calculate_paw_features()
    for a in range(len(ae_feat)):
        wt = get_atom_wt(functional, a)
        for feat, sgn in [(ae_feat[a], 1), (ps_feat[a], -1)]:
            assert wt.size == feat.shape[-1]
            yield feat, sgn * wt


def get_atom_feat_wt_deriv(functional, DD_aop, p_o):
    ae_feat, ps_feat, ae_dfeat, ps_dfeat = functional.calculate_paw_features_deriv(
        DD_aop, p_o
    )
    for a in range(len(ae_feat)):
        lst = [(ae_feat[a], ae_dfeat[a], 1), (ps_feat[a], ps_dfeat[a], -1)]
        wt = get_atom_wt(functional, a)
        for feat, dfeat, sgn in lst:
            assert wt.size == feat.shape[-1] == dfeat.shape[-1]
            yield feat, dfeat, sgn * wt


def get_features_and_weights(calc, xc, screen_dens=True):
    vol = calc.atoms.get_volume()
    ae_feat = xc.get_features_on_grid()
    nspin, nfeat = ae_feat.shape[:2]
    size = np.prod(ae_feat.shape[2:])
    all_wt = np.ones(size) / size * vol / Bohr**3
    all_wt = all_wt.flatten()
    ae_feat = ae_feat.reshape(nspin, nfeat, -1)
    # hopefully save disk space and avoid negative density
    if screen_dens:
        cond = ae_feat[:, 0, :].sum(axis=0) > 1e-6
        ae_feat = ae_feat[..., cond]
        all_wt = all_wt[cond]
    assert all_wt.size == ae_feat.shape[-1]

    if xc.has_paw:
        for ft, wt in get_atom_feat_wt(xc):
            assert wt.shape[-1] == ft.shape[-1]
            ae_feat = np.append(ae_feat, ft, axis=-1)
            all_wt = np.append(all_wt, wt, axis=-1)
    return ae_feat, all_wt


def get_features_and_weights_deriv(
    calc, xc, p_j, drhodf_jxg, DD_aop=None, screen_dens=True
):
    vol = calc.atoms.get_volume()
    feat_sig, dfeat_jig = xc.get_features_on_grid_deriv(p_j, drhodf_jxg, DD_aop=DD_aop)
    nspin, nfeat = feat_sig.shape[:2]
    size = np.prod(feat_sig.shape[2:])
    all_wt = np.ones(size) / size * vol / Bohr**3
    all_wt = all_wt.flatten()
    feat_sig = feat_sig.reshape(nspin, nfeat, -1)
    dfeat_jig = dfeat_jig.reshape(len(p_j), nfeat, -1)
    # hopefully save disk space and avoid negative density
    if screen_dens:
        cond = feat_sig[:, 0, :].sum(axis=0) > 1e-6
        feat_sig = feat_sig[..., cond]
        dfeat_jig = dfeat_jig[..., cond]
        all_wt = all_wt[cond]
    assert all_wt.size == feat_sig.shape[-1]

    if xc.has_paw:
        assert DD_aop is not None
        DD_aop = xc.atomdist.to_work(DD_aop)
        for ft, dft, wt in get_atom_feat_wt_deriv(xc, DD_aop, p_j):
            assert wt.shape[-1] == ft.shape[-1]
            feat_sig = np.append(feat_sig, ft, axis=-1)
            dfeat_jig = np.append(dfeat_jig, dft, axis=-1)
            all_wt = np.append(all_wt, wt, axis=-1)
    return feat_sig, dfeat_jig, all_wt


def get_drho_df(calc, p_i):
    dens = calc.density
    nspin = dens.nt_sg.shape[0]
    wfs = calc.wfs
    assert isinstance(dens, ReciprocalSpaceDensity)
    shape = (len(p_i), 4)
    real_ixR = wfs.gd.zeros(shape)
    imag_ixR = wfs.gd.zeros(shape)
    drhodf_ixR = wfs.gd.zeros((len(p_i), 5))
    for i, (s, k, n) in enumerate(p_i):
        rank, q = wfs.kd.who_has(k)
        if wfs.kd.comm.rank == rank:
            u = q * nspin + s
            kpt = wfs.kpt_u[u]
            assert kpt.s == s
            assert kpt.k == k
        else:
            continue
        G_Gv = wfs.pd.get_reciprocal_vectors(q=kpt.q)
        psi_G = kpt.psit_nG[n]
        psi_R = wfs.pd.ifft(psi_G, kpt.q)
        real_ixR[i, 0] = psi_R.real
        imag_ixR[i, 0] = psi_R.imag
        for v in range(3):
            psi_G = 1j * G_Gv[:, v] * kpt.psit_nG[n]
            psi_R = wfs.pd.ifft(psi_G, kpt.q)
            real_ixR[i, v + 1] = psi_R.real
            imag_ixR[i, v + 1] = psi_R.imag
        drhodf_ixR[i, 0] = (
            real_ixR[i, 0] * real_ixR[i, 0] + imag_ixR[i, 0] * imag_ixR[i, 0]
        )
        for v in range(1, 4):
            drhodf_ixR[i, v] = 2 * (
                real_ixR[i, 0] * real_ixR[i, v] + imag_ixR[i, 0] * imag_ixR[i, v]
            )
            drhodf_ixR[i, 4] += 0.5 * (
                real_ixR[i, v] * real_ixR[i, v] + imag_ixR[i, v] * imag_ixR[i, v]
            )
    wfs.kptband_comm.sum(drhodf_ixR)
    # for i in range(len(p_i)):
    #    for x in range(5):
    #        wfs.kd.symmetry.symmetrize(drhodf_ixR[i, x], wfs.gd)
    return drhodf_ixR


def interpolate_drhodf(gd, interp, drhodf_ixR):
    drhodf_ixg = gd.zeros(drhodf_ixR.shape[:2])
    for i in range(drhodf_ixR.shape[0]):
        for x in range(drhodf_ixR.shape[1]):
            interp(drhodf_ixR[i, x], drhodf_ixg[i, x])
    return drhodf_ixg


class _FeatureMixin:
    def get_D_asp(self):
        return self.atomdist.to_work(self.dens.D_asp)

    def initialize_paw_kernel(self, cider_kernel_inp, Nalpha_atom, encut_atom):
        self.paw_kernel = PASDWCiderFeatureKernel(
            cider_kernel_inp,
            self._plan,
            self.gd,
            self.cut_xcgrid,
        )
        self.paw_kernel.initialize(
            self.dens, self.atomdist, self.atom_partition, self.setups
        )
        self.paw_kernel.initialize_more_things(self.setups)
        self.paw_kernel.interpolate_dn1 = self.interpolate_dn1
        self.atom_slices_s = None

    def _collect_feat(self, feat_xg):
        # TODO this has some redundant communication
        xshape = feat_xg.shape[:-3]
        gshape_out = tuple(self.gd.n_c)
        _feat_xg = np.zeros(xshape + gshape_out)
        self._add_from_cider_grid(_feat_xg, feat_xg)
        feat_xg = self.gd.collect(_feat_xg, broadcast=True)
        feat_xg.shape = xshape + (-1,)
        return feat_xg

    def _get_features_on_grid(self, rho_sxg):
        nspin = len(rho_sxg)
        self.nspin = nspin
        self._DD_aop = None
        if not self.is_initialized:
            self._setup_plan()
            self.fft_obj.initialize_backend()
            self.initialize_more_things()
        plan = self._plan
        rho_tuple = plan.get_rho_tuple(rho_sxg, with_spin=True)
        feat_sig = self.call_fwd(rho_tuple)[0]
        # TODO version i features
        feat_sig[:] *= plan.nspin
        return feat_sig

    def _get_pseudo_density(self):
        nt_sg = self.dens.nt_sg
        redist = self._hamiltonian.xc_redistributor
        if redist is not None:
            nt_sg = redist.distribute(nt_sg)
        return nt_sg

    def get_features_on_grid(self):
        nt_sg = self._get_pseudo_density()
        taut_sg = self._get_taut(nt_sg)[0]
        self.timer.start("Reorganize density")
        rho_sxg = self._get_cider_inputs(nt_sg, taut_sg)[1]
        self.timer.stop()
        feat = self._get_features_on_grid(rho_sxg)
        return self._collect_feat(feat)

    def _set_paw_terms(self):
        pass

    def _get_features_on_grid_deriv(self, p_j, rho_sxg, drhodf_jxg, DD_aop=None):
        nspin = len(rho_sxg)
        self.nspin = nspin
        self._p_o = p_j
        if not self.is_initialized:
            self._setup_plan()
            self.fft_obj.initialize_backend()
            self.initialize_more_things()
        if self.has_paw:
            self._DD_aop = self.atomdist.to_work(DD_aop)
        else:
            self._DD_aop = None
        plan = self._plan
        rho_tuple = plan.get_rho_tuple(rho_sxg, with_spin=True)
        drhodf_j_tuple = []
        for j, drhodf_xg in enumerate(drhodf_jxg):
            drhodf_j_tuple.append(plan.get_drhodf_tuple(rho_sxg[p_j[j][0]], drhodf_xg))
        feat_sig, dfeat_sig = self.call_fwd(rho_tuple)[:2]
        # TODO version i features
        feat_sig[:] *= plan.nspin
        dfeatdf_jig = []
        for j, drhodf_tuple in enumerate(drhodf_j_tuple):
            s = p_j[j][0]
            _rho_tuple = tuple([r[s] for r in rho_tuple])
            dfeatdf_ig = self.call_feat_deriv(j, _rho_tuple, drhodf_tuple, dfeat_sig[s])
            dfeatdf_jig.append(dfeatdf_ig)
        dfeatdf_jig = np.ascontiguousarray(dfeatdf_jig)
        dfeatdf_jig[:] *= plan.nspin

        return feat_sig, dfeatdf_jig

    def get_features_on_grid_deriv(self, p_j, drhodf_jxg, DD_aop=None):
        """
        Get the pseudo-features on the uniform FFT grid used
        to integrate the XC energy, along with the derivatives
        of the features with respect to orbital occupation.

        Args:
            self: CiderFunctional
            p_j: List of (s, k, n) band indexes
            drhodf_jxg: (norb, 5, ngrid):
                Derivatives of the density (0), gradient (1,2,3),
                and kinetic energy density (4) with respect to orbital
                occupation number.
            DD_aop (dict of numpy array): derivatives of atomic density
                matrices with respect to orbital occupation.

        Returns:
            feat_sig, dfeatdf_jig
        """
        nt_sg = self._get_pseudo_density()
        taut_sg = self._get_taut(nt_sg)[0]
        self.timer.start("Reorganize density")
        rho_sxg = self._get_cider_inputs(nt_sg, taut_sg)[1]
        drhodf_jxg = self._distribute_to_cider_grid(drhodf_jxg)
        self.timer.stop()
        nx = rho_sxg.shape[1]
        feat, dfeat = self._get_features_on_grid_deriv(
            p_j, rho_sxg, drhodf_jxg[:, :nx], DD_aop=DD_aop
        )
        return self._collect_feat(feat), self._collect_feat(dfeat)

    def communicate_paw_features(self, ae_feat, ps_feat, ncomponents, nfeat):
        assert len(ae_feat) == len(ps_feat)
        feat_shapes = [
            (ncomponents, nfeat, Y_nL.shape[0] * setup.xc_correction.rgd.N)
            for setup in self.setups
        ]
        ae_feat_glob = self.atom_partition.arraydict(feat_shapes, float)
        ps_feat_glob = self.atom_partition.arraydict(feat_shapes, float)
        for a in list(ae_feat.keys()):
            ae_feat_glob[a] = ae_feat[a]
            ps_feat_glob[a] = ps_feat[a]
        dist = self.atomdist
        ae_feat_glob = dist.from_work(ae_feat_glob)
        ps_feat_glob = dist.from_work(ps_feat_glob)
        ae_feat_glob = {
            a: ae_feat_glob[a] for a in ae_feat_glob
        }  # turn to normal array
        ps_feat_glob = {
            a: ps_feat_glob[a] for a in ps_feat_glob
        }  # turn to normal array
        atom_comm = dist.comm
        rank_a = dist.partition.rank_a
        for a in range(len(self.setups)):
            if rank_a[a] == 0:
                if atom_comm.rank == 0:
                    assert a in ae_feat_glob
            else:
                src = rank_a[a]
                if atom_comm.rank == src:
                    atom_comm.send(ae_feat_glob[a], 0, tag=2 * a)
                    atom_comm.send(ps_feat_glob[a], 0, tag=2 * a + 1)
                elif atom_comm.rank == 0:
                    ae_feat_glob[a] = np.empty(feat_shapes[a])
                    ps_feat_glob[a] = np.empty(feat_shapes[a])
                    atom_comm.receive(ae_feat_glob[a], src, tag=2 * a)
                    atom_comm.receive(ps_feat_glob[a], src, tag=2 * a + 1)
            if atom_comm.rank != 0:
                ae_feat_glob[a] = np.empty(feat_shapes[a])
                ps_feat_glob[a] = np.empty(feat_shapes[a])
            atom_comm.broadcast(ae_feat_glob[a], 0)
            atom_comm.broadcast(ps_feat_glob[a], 0)
        return ae_feat_glob, ps_feat_glob

    def call_feat_deriv(self, j, rho_tuple, drhodf_tuple, dfeat_ig):
        p, dp = None, None
        plan = self._plan
        arg_g, darg_g = plan.get_interpolation_arguments(rho_tuple, i=-1)
        fun_g, dfun_g = plan.get_function_to_convolve(rho_tuple)
        dargdf_g = 0
        dfundf_g = 0
        for darg, dfun, drho in zip(darg_g, dfun_g, drhodf_tuple):
            dargdf_g += darg * drho
            dfundf_g += dfun * drho
        p, dp = plan.get_interpolation_coefficients(arg_g, i=-1, vbuf=p, dbuf=dp)
        p.shape = arg_g.shape + (p.shape[-1],)
        dp.shape = arg_g.shape + (p.shape[-1],)
        dthetadf = dfundf_g[..., None] * p + (fun_g * dargdf_g)[..., None] * dp
        self.fft_obj.set_work(np.ones_like(fun_g), dthetadf)
        self.set_fft_work(self.nspin + j)
        self.fft_obj.compute_forward_convolution()
        self.get_fft_work(self.nspin + j)
        shape = (plan.nldf_settings.nfeat,) + arg_g.shape
        dfeatdf_ig = np.zeros(shape)
        for i in range(plan.num_vj):
            a_g, da_g = plan.get_interpolation_arguments(rho_tuple, i=i)
            p, dp = plan.get_interpolation_coefficients(a_g, i=i, vbuf=p, dbuf=dp)
            if False:  # TODO might need this later. Or can it be removed?
                coeff_multipliers = None
                p[:] *= coeff_multipliers
                dp[:] *= coeff_multipliers
            if False:  # TODO might need this later
                plan.get_transformed_interpolation_terms(p, i=i, fwd=True, inplace=True)
                plan.get_transformed_interpolation_terms(
                    dp, i=i, fwd=True, inplace=True
                )
            p.shape = a_g.shape + (plan.nalpha,)
            self.fft_obj.fill_vj_feature_(dfeatdf_ig[i], p)
            dadf_g = 0
            for darg, drho in zip(da_g, drhodf_tuple):
                dadf_g += darg * drho
            dfeatdf_ig[i] += dadf_g * dfeat_ig[i]
        return dfeatdf_ig

    def calculate_paw_features(self):
        ae_feat, ps_feat = self.paw_kernel.calculate_paw_cider_features_p2(
            self.setups, self.D_asp, self.c_asiq, None, None
        )
        nfeat = self.cider_kernel.mlfunc.settings.nldf_settings.nfeat
        return self.communicate_paw_features(ae_feat, ps_feat, self.nspin, nfeat)

    def calculate_paw_features_deriv(self, DD_aop, p_o):
        res = self.paw_kernel.calculate_paw_cider_features_p2(
            self.setups, self.D_asp, self.c_asiq, DD_aop, p_o
        )
        ae_feat, ps_feat, ae_dfeat, ps_dfeat = res
        nfeat = self.cider_kernel.mlfunc.settings.nldf_settings.nfeat
        ae_feat, ps_feat = self.communicate_paw_features(
            ae_feat, ps_feat, self.nspin, nfeat
        )
        ae_dfeat, ps_dfeat = self.communicate_paw_features(
            ae_dfeat, ps_dfeat, len(p_o), nfeat
        )
        return ae_feat, ps_feat, ae_dfeat, ps_dfeat


class _PAWFeatureMixin(_FeatureMixin):
    def _set_paw_terms(self):
        if not hasattr(self, "setups") or self.setups is None:
            return
        if self._DD_aop is None:
            D_asp = self.get_D_asp()
            if not (D_asp.partition.rank_a == self.atom_partition.rank_a).all():
                raise ValueError("rank_a mismatch")
            self.c_asiq, self.df_asgLq = self.paw_kernel.calculate_paw_feat_corrections(
                self.setups, D_asp
            )
            self.D_asp = D_asp
        else:
            D_asp = self.get_D_asp()
            if not (D_asp.partition.rank_a == self.atom_partition.rank_a).all():
                raise ValueError("rank_a mismatch")
            res = self.paw_kernel.calculate_paw_cider_features_p1(
                self.setups, D_asp, self._DD_aop, self._p_o
            )
            self.c_asiq, self.df_asgLq = res
            self.D_asp = D_asp


class _SLFeatMixin(_FeatureMixin):
    def calculate_paw_features(self):
        settings = self.cider_kernel.mlfunc.settings.sl_settings
        ae_feat, ps_feat = calculate_paw_sl_features(
            self.setups, self.get_D_asp(), settings.level == "MGGA"
        )
        if settings.mode != "l":
            plan = SemilocalPlan(settings, self.nspin)
            for a in list(ae_feat.keys()):
                ae_feat[a] = plan.get_feat(ae_feat[a])
                ps_feat[a] = plan.get_feat(ps_feat[a])
        return self.communicate_paw_features(
            ae_feat, ps_feat, self.nspin, settings.nfeat
        )

    def calculate_paw_features_deriv(self, DD_aop, p_o):
        settings = self.cider_kernel.mlfunc.settings.sl_settings
        _ae_rho, _ps_rho, _ae_drho, _ps_drho = calculate_paw_sl_features_deriv(
            self.setups, self.get_D_asp(), DD_aop, p_o, settings.level == "MGGA"
        )

        if settings.mode != "l":
            ae_feat = {}
            ae_dfeat = {}
            ps_feat = {}
            ps_dfeat = {}
            plan = SemilocalPlan(settings, self.nspin)
            for a in list(_ae_rho.keys()):
                ae_drho = _ae_drho[a]
                ae_rho = _ae_rho[a]
                ps_drho = _ps_drho[a]
                ps_rho = _ps_rho[a]
                ae_dfeat[a] = np.empty((len(p_o), settings.nfeat) + ae_drho.shape[2:])
                for j, p in enumerate(p_o):
                    ae_dfeat[a][j : j + 1] = plan.get_occd(
                        ae_rho[p[0] : p[0] + 1], ae_drho[j : j + 1]
                    )[1]
                ae_feat[a] = plan.get_feat(ae_rho)
                ps_dfeat[a] = np.empty((len(p_o), settings.nfeat) + ps_drho.shape[2:])
                for j, p in enumerate(p_o):
                    ps_dfeat[a][j : j + 1] = plan.get_occd(
                        ps_rho[p[0] : p[0] + 1], ps_drho[j : j + 1]
                    )[1]
                ps_feat[a] = plan.get_feat(ps_rho)
        else:
            ae_feat = _ae_rho
            ae_dfeat = _ae_drho
            ps_feat = _ps_rho
            ps_dfeat = _ps_drho

        ae_feat, ps_feat = self.communicate_paw_features(
            ae_feat, ps_feat, self.nspin, settings.nfeat
        )
        ae_dfeat, ps_dfeat = self.communicate_paw_features(
            ae_dfeat, ps_dfeat, len(p_o), settings.nfeat
        )
        return ae_feat, ps_feat, ae_dfeat, ps_dfeat

    def _collect_feat(self, feat_xg):
        # TODO this has some redundant communication
        xshape = feat_xg.shape[:-1]
        gshape_out = tuple(self.gd.n_c)
        gshape_in = tuple(self.distribution.local_output_size_c)
        feat_xg = feat_xg.view()
        feat_xg.shape = xshape + gshape_in
        _feat_xg = np.zeros(xshape + gshape_out)
        self._add_from_cider_grid(_feat_xg, feat_xg)
        feat_xg = self.gd.collect(_feat_xg, broadcast=True)
        feat_xg.shape = xshape + (-1,)
        return feat_xg

    def get_features_on_grid(self):
        nt_sg = self._get_pseudo_density()
        taut_sg = self._get_taut(nt_sg)[0]
        rho_sxg = self._get_cider_inputs(nt_sg, taut_sg)[1]
        settings = self.cider_kernel.mlfunc.settings.sl_settings
        rho_sxg = rho_sxg.view()
        rho_sxg.shape = rho_sxg.shape[:2] + (-1,)
        if settings.mode != "l":
            plan = SemilocalPlan(settings, self.nspin)
            feat_sig = plan.get_feat(rho_sxg)
        else:
            feat_sig = rho_sxg
        return self._collect_feat(feat_sig)

    def get_features_on_grid_deriv(self, p_j, drhodf_jxg, DD_aop=None):
        nt_sg = self._get_pseudo_density()
        taut_sg = self._get_taut(nt_sg)[0]
        self.timer.start("Reorganize density")
        rho_sxg = self._get_cider_inputs(nt_sg, taut_sg)[1]
        drhodf_jxg = self._distribute_to_cider_grid(drhodf_jxg)
        self.timer.stop()
        rho_sxg = rho_sxg.view()
        drhodf_jxg = drhodf_jxg.view()
        rho_sxg.shape = rho_sxg.shape[:2] + (-1,)
        drhodf_jxg.shape = drhodf_jxg.shape[:2] + (-1,)
        settings = self.cider_kernel.mlfunc.settings.sl_settings
        if settings.mode != "l":
            plan = SemilocalPlan(settings, self.nspin)
            dfeat_jig = np.empty((len(p_j), settings.nfeat) + rho_sxg.shape[2:])
            for j, p in enumerate(p_j):
                dfeat_jig[j : j + 1] = plan.get_occd(
                    rho_sxg[p[0] : p[0] + 1], drhodf_jxg[j : j + 1]
                )[1]
            feat_sig = plan.get_feat(rho_sxg)
        else:
            feat_sig = rho_sxg
            dfeat_jig = drhodf_jxg
        return self._collect_feat(feat_sig), self._collect_feat(dfeat_jig)


def _initialize_gga_feat(self, density, hamiltonian, wfs):
    """
    GGAs do not automatically define these distribute and restrict
    functions, which we need for feature generation, so this initializer
    takes care of that.
    """
    super(self.__class__, self).initialize(density, hamiltonian, wfs)
    self.wfs = wfs
    self.atomdist = hamiltonian.atomdist
    self.setups = hamiltonian.setups
    self.atom_partition = self.get_D_asp().partition
    self._hamiltonian = hamiltonian
    if (not hasattr(hamiltonian, "xc_redistributor")) or (
        hamiltonian.xc_redistributor is None
    ):
        self.restrict_and_collect = hamiltonian.restrict_and_collect
        self.distribute_and_interpolate = density.distribute_and_interpolate
    else:

        def _distribute_and_interpolate(in_xR, out_xR=None):
            tmp_xR = density.interpolate(in_xR)
            if hamiltonian.xc_redistributor.enabled:
                out_xR = hamiltonian.xc_redistributor.distribute(tmp_xR, out_xR)
            elif out_xR is None:
                out_xR = tmp_xR
            else:
                out_xR[:] = tmp_xR
            return out_xR

        def _restrict_and_collect(in_xR, out_xR=None):
            if hamiltonian.xc_redistributor.enabled:
                in_xR = hamiltonian.xc_redistributor.collect(in_xR)
            return hamiltonian.restrict(in_xR, out_xR)

        self.restrict_and_collect = _restrict_and_collect
        self.distribute_and_interpolate = _distribute_and_interpolate


class FeatGGA(_FeatureMixin, CiderGGA):
    initialize = _initialize_gga_feat


class FeatMGGA(_FeatureMixin, CiderMGGA):
    initialize = _initialize_gga_feat


class FeatGGAPAW(_PAWFeatureMixin, CiderGGAPASDW):
    initialize = _initialize_gga_feat


class FeatMGGAPAW(_PAWFeatureMixin, CiderMGGAPASDW):
    pass


class FeatSLGGA(_SLFeatMixin, CiderGGA):
    initialize = _initialize_gga_feat


class FeatSLMGGA(_SLFeatMixin, CiderMGGA):
    initialize = _initialize_gga_feat


class FeatSLGGAPAW(_SLFeatMixin, CiderGGAPASDW):
    initialize = _initialize_gga_feat


class FeatSLMGGAPAW(_SLFeatMixin, CiderMGGAPASDW):
    initialize = _initialize_gga_feat



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import numpy as np
from pyscf import gto
from pyscf.data.radii import COVALENT as COVALENT_RADII

from ciderpress.dft.lcao_convolutions import (
    DEFAULT_AUX_BETA,
    DEFAULT_CIDER_LMAX,
    ATCBasis,
    ConvolutionCollection,
    ConvolutionCollectionK,
    get_convolution_expnts_from_expnts,
    get_etb_from_expnt_range,
    get_gamma_lists_from_bas_and_env,
)
from ciderpress.dft.lcao_interpolation import LCAOInterpolator, LCAOInterpolatorDirect
from ciderpress.dft.lcao_nldf_generator import LCAONLDFGenerator
from ciderpress.dft.plans import NLDFGaussianPlan, NLDFSplinePlan, get_ccl_settings


def aug_etb_for_cider(
    mol,
    beta=DEFAULT_AUX_BETA,
    upper_fac=1.0,
    lower_fac=1.0,
    lmax=DEFAULT_CIDER_LMAX,
    def_afac_max=4.0,
    def_afac_min=0.5,
):
    """
    augment weigend basis with even-tempered gaussian basis
    exps = alpha*beta^i for i = 1..N
    For this implementation, exponent values should be on the
    same 'grid' ..., 1/beta, 1, beta, beta^2, ...

    Args:
        mol (pyscf.gto.Mole): molecule object
        beta (float): exponent parameter for ETB basis
        upper_fac (float): Max exponent should be at least
            upper_fac * max exponent in original basis
            for each l value.
        lower_fac (float): Min exponent should be at most
            lower_fac * min exponent in original basis
            for each l value.
        lmax (int): Maximum l value for expansion.
        def_afac_max (float): Max exponent relative to Bohr radius
            if no orbital products with a particular l exist
            in the original basis.
        def_afac_min (float): Min exponent relative to Bohr radius
            if not orbital products with a particular l exist
            in the original basis

    Returns:
        New basis set for expanding CIDER theta contributions.
    """
    uniq_atoms = set([a[0] for a in mol._atom])

    newbasis = {}
    for symb in uniq_atoms:
        nuc_charge = gto.charge(symb)

        def_amax = def_afac_max / COVALENT_RADII[nuc_charge] ** 2
        def_amin = def_afac_min / COVALENT_RADII[nuc_charge] ** 2

        max_shells = lmax
        emin_by_l = [1e99] * 8
        emax_by_l = [0] * 8
        for b in mol._basis[symb]:
            l = b[0]
            if l >= max_shells + 1:
                continue

            if isinstance(b[1], int):
                e_c = np.array(b[2:])
            else:
                e_c = np.array(b[1:])
            es = e_c[:, 0]
            cs = e_c[:, 1:]
            es = es[abs(cs).max(axis=1) > 1e-3]
            emax_by_l[l] = max(es.max(), emax_by_l[l])
            emin_by_l[l] = min(es.min(), emin_by_l[l])

        etb = get_etb_from_expnt_range(
            lmax,
            beta,
            emin_by_l,
            emax_by_l,
            def_amax,
            def_amin,
            lower_fac=lower_fac,
            upper_fac=upper_fac,
        )
        newbasis[symb] = gto.expand_etbs(etb)
    return newbasis


def get_gamma_lists_from_mol(mol):
    return get_gamma_lists_from_bas_and_env(mol._bas, mol._env)


class PySCFNLDFInitializer:
    def __init__(self, settings, **kwargs):
        """
        Args:
            settings (NLDFSettings): settings for features
            **kwargs: Any of the optional arguments in the
                PyscfNLDFGenerator.from_mol_and_settings
                function.
        """
        self.settings = settings
        self.kwargs = kwargs

    def initialize_nldf_generator(self, mol, grids_indexer, nspin):
        return PyscfNLDFGenerator.from_mol_and_settings(
            mol, grids_indexer, nspin, self.settings, **self.kwargs
        )


class PyscfNLDFGenerator(LCAONLDFGenerator):
    """
    A PySCF-specific wrapper for the NLDF feature generator.
    """

    @classmethod
    def from_mol_and_settings(
        cls,
        mol,
        grids_indexer,
        nspin,
        nldf_settings,
        plan_type="gaussian",
        lmax=None,
        aux_lambd=1.6,
        aug_beta=None,
        alpha_max=10000,
        alpha_min=None,
        alpha_formula=None,
        rhocut=1e-10,
        expcut=1e-10,
        gbuf=2.0,
        interpolator_type="onsite_direct",
        aparam=0.03,
        dparam=0.04,
        nrad=200,
    ):
        if lmax is None:
            lmax = grids_indexer.lmax
        if lmax > grids_indexer.lmax:
            raise ValueError("lmax cannot be larger than grids_indexer.lmax")
        if interpolator_type not in ["onsite_direct", "onsite_spline", "train_gen"]:
            raise ValueError
        if aug_beta is None:
            aug_beta = aux_lambd
        if alpha_min is None:
            alpha_min = nldf_settings.theta_params[0] / 256  # sensible default
        if plan_type not in ["gaussian", "spline"]:
            raise ValueError("plan_type must be gaussian or spline")
        if alpha_formula is None:
            alpha_formula = "etb" if plan_type == "gaussian" else "zexp"
        nalpha = int(np.ceil(np.log(alpha_max / alpha_min) / np.log(aux_lambd))) + 1
        plan_class = NLDFGaussianPlan if plan_type == "gaussian" else NLDFSplinePlan
        plan = plan_class(
            nldf_settings,
            nspin,
            alpha_min,
            aux_lambd,
            nalpha,
            coef_order="gq",
            alpha_formula=alpha_formula,
            proc_inds=None,
            rhocut=rhocut,
            expcut=expcut,
        )
        basis = aug_etb_for_cider(mol, lmax=lmax, beta=aug_beta)
        mol = gto.M(
            atom=mol.atom,
            basis=basis,
            spin=mol.spin,
            charge=mol.charge,
            unit=mol.unit,
        )
        dat = get_gamma_lists_from_mol(mol)
        atco_inp = ATCBasis(*dat)
        dat = get_convolution_expnts_from_expnts(
            plan.alphas, dat[0], dat[1], dat[2], dat[4], gbuf=gbuf
        )
        atco_out = ATCBasis(*dat)
        if plan.nldf_settings.nldf_type == "k":
            ccl = ConvolutionCollectionK(
                atco_inp, atco_out, plan.alphas, plan.alpha_norms
            )
        else:
            has_vj, ifeat_ids = get_ccl_settings(plan)
            ccl = ConvolutionCollection(
                atco_inp,
                atco_out,
                plan.alphas,
                plan.alpha_norms,
                has_vj=has_vj,
                ifeat_ids=ifeat_ids,
            )
        if interpolator_type == "train_gen":
            interpolator = LCAOInterpolator.from_ccl(
                mol.atom_coords(unit="Bohr"),
                ccl,
                aparam=aparam,
                dparam=dparam,
                nrad=nrad,
            )
        else:
            onsite_direct = interpolator_type == "onsite_direct"
            interpolator = LCAOInterpolatorDirect.from_ccl(
                grids_indexer,
                mol.atom_coords(unit="Bohr"),
                ccl,
                aparam=aparam,
                dparam=dparam,
                nrad=nrad,
                onsite_direct=onsite_direct,
            )
        ccl.compute_integrals_()
        ccl.solve_projection_coefficients()
        return cls(plan, ccl, interpolator, grids_indexer)

    def get_extra_ao(self):
        return (self.interpolator.nlm + 4) * self.natm



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

from pathlib import Path

import numpy as np
from ase.units import Ha
from gpaw import GPAW
from gpaw.calculator import GPAW as GPAWOld
from gpaw.new.ase_interface import ASECalculator
from gpaw.utilities import unpack_hermitian
from gpaw.xc import XC
from gpaw.xc.kernel import XCNull

from ciderpress.gpaw.cider_paw import CiderPASDW_MPRoutines


def non_self_consistent_eigenvalues(
    calc, xc, n1=0, n2=0, kpt_indices=None, snapshot=None, get_ham=False
):
    if not isinstance(calc, (GPAWOld, ASECalculator)):
        if calc == "<gpw-file>":  # for doctest
            return (np.zeros((1, 1, 1)), np.zeros((1, 1, 1)), np.zeros((1, 1, 1)))
        calc = GPAW(Path(calc), txt=None, parallel={"band": 1, "kpt": 1})

    assert isinstance(calc, (GPAWOld, ASECalculator))
    wfs = calc.wfs

    if isinstance(xc, str):
        xc = XC(xc)

    if n2 <= 0:
        n2 += wfs.bd.nbands

    if kpt_indices is None:
        kpt_indices = np.arange(wfs.kd.nibzkpts).tolist()

    if snapshot is not None:
        raise NotImplementedError
    # path = Path(snapshot) if snapshot is not None else None

    e_dft_sin = np.zeros(0)
    v_dft_sin = np.zeros(0)

    # sl=semilocal, nl=nonlocal
    v_hyb_sl_sin = np.zeros(0)

    if v_dft_sin.size == 0:
        e_dft_sin, v_dft_sin, v_hyb_sl_sin = _semi_local(
            calc, xc, n1, n2, kpt_indices, get_ham=get_ham
        )

    return (e_dft_sin * Ha, v_dft_sin * Ha, v_hyb_sl_sin * Ha)


def _semi_local(calc, xc, n1, n2, kpt_indices, get_ham=False):
    wfs = calc.wfs
    nspins = wfs.nspins
    e_dft_sin = np.array(
        [
            [calc.get_eigenvalues(k, spin)[n1:n2] for k in kpt_indices]
            for spin in range(nspins)
        ]
    )
    gs = calc.gs_adapter()
    for a in range(len(calc.setups)):
        setup = calc.setups[a]
        pawdata = gs.pawdatasets.by_atom[a]
        if setup.xc_correction is not None:
            pawdata.xc_correction.dc_g = setup.xc_correction.dc_g
            pawdata.xc_correction.dct_g = setup.xc_correction.dct_g
            pawdata.xc_correction.d_qg = setup.xc_correction.d_qg
            pawdata.xc_correction.dt_qg = setup.xc_correction.dt_qg
            pawdata.xc_correction.tau_npg = setup.xc_correction.tau_npg
            pawdata.xc_correction.taut_npg = setup.xc_correction.taut_npg
            pawdata.xc_correction.tauc_g = setup.xc_correction.tauc_g
            pawdata.xc_correction.tauct_g = setup.xc_correction.tauct_g
    if get_ham:
        v_dft_sin = vxc_mat(gs, n1=n1, n2=n2)[:, kpt_indices]
    else:
        v_dft_sin = vxc(gs, n1=n1, n2=n2)[:, kpt_indices]
    if isinstance(xc.kernel, XCNull):
        v_hyb_sl_sin = np.zeros_like(v_dft_sin)
    else:
        gs = calc.gs_adapter()
        for a in range(len(calc.setups)):
            setup = calc.setups[a]
            pawdata = gs.pawdatasets.by_atom[a]
            if setup.xc_correction is not None:
                pawdata.xc_correction.dc_g = setup.xc_correction.dc_g
                pawdata.xc_correction.dct_g = setup.xc_correction.dct_g
                pawdata.xc_correction.d_qg = setup.xc_correction.d_qg
                pawdata.xc_correction.dt_qg = setup.xc_correction.dt_qg
                pawdata.xc_correction.tau_npg = setup.xc_correction.tau_npg
                pawdata.xc_correction.taut_npg = setup.xc_correction.taut_npg
                pawdata.xc_correction.tauc_g = setup.xc_correction.tauc_g
                pawdata.xc_correction.tauct_g = setup.xc_correction.tauct_g
        if get_ham:
            v_hyb_sl_sin = vxc_mat(gs, xc, n1=n1, n2=n2)[:, kpt_indices]
        else:
            v_hyb_sl_sin = vxc(gs, xc, n1=n1, n2=n2)[:, kpt_indices]
    return e_dft_sin / Ha, v_dft_sin / Ha, v_hyb_sl_sin / Ha


def vxc(gs, xc=None, coredensity=True, n1=0, n2=0):
    """Calculate XC-contribution to eigenvalues."""

    if n2 <= 0:
        n2 += gs.bd.nbands

    ham = gs.hamiltonian
    dens = gs.density

    if xc is None:
        xc = ham.xc
    elif isinstance(xc, str):
        xc = XC(xc)

    if dens.nt_sg is None:
        dens.interpolate_pseudo_density()

    gs_gd = gs._hamiltonian.xc.gd
    xc.set_grid_descriptor(gs_gd)
    xc.initialize(gs._density, gs._hamiltonian, gs._wfs)
    xc.set_positions(gs.spos_ac)

    # Calculate XC-potential
    vxct_sg = ham.finegd.zeros(gs.nspins)
    redist = gs._hamiltonian.xc_redistributor
    if redist is not None:
        nt_sg = redist.distribute(dens.nt_sg)
        _vxc = redist.distribute(vxct_sg)
    else:
        nt_sg = dens.nt_sg
        _vxc = vxct_sg
    xc.calculate(gs_gd, nt_sg, _vxc)
    if redist is not None:
        redist.collect(_vxc, vxct_sg)
    vxct_sG = ham.restrict_and_collect(vxct_sg)

    # Calculate PAW corrections:
    dvxc_asii = {}
    if isinstance(xc, CiderPASDW_MPRoutines):
        xc._collect_paw_corrections()
    for a, D_sp in dens.D_asp.items():
        dvxc_sp = np.zeros_like(D_sp)

        pawdata = gs.pawdatasets.by_atom[a]
        xc.calculate_paw_correction(
            pawdata, D_sp, dvxc_sp, a=a, addcoredensity=coredensity
        )
        dvxc_asii[a] = [unpack_hermitian(dvxc_p) for dvxc_p in dvxc_sp]

    vxc_un = np.empty((gs.kd.mynk * gs.nspins, gs.bd.mynbands))
    for u, vxc_n in enumerate(vxc_un):
        kpt = gs.kpt_u[u]
        vxct_G = vxct_sG[kpt.s]
        if xc.type == "MGGA":
            G_Gv = gs._wfs.pd.get_reciprocal_vectors(q=kpt.q)
        for n in range(gs.bd.mynbands):
            if n1 <= n + gs.bd.beg < n2:
                psit_G = gs.get_wave_function_array(u, n)
                vxc_n[n] = gs.gd.integrate(
                    (psit_G * psit_G.conj()).real, vxct_G, global_integral=False
                )
                if xc.type == "MGGA":
                    psit_k = kpt.psit_nG[n]
                    dpsit2_G = np.zeros_like(vxct_G)
                    for v in range(3):
                        tmp_G = gs.pd.ifft(1j * G_Gv[:, v] * psit_k, q=kpt.q)
                        dpsit2_G += 0.5 * (tmp_G * tmp_G.conj()).real
                    vxc_n[n] += gs.gd.integrate(
                        dpsit2_G, xc.dedtaut_sG[kpt.s], global_integral=False
                    )

        for a, dvxc_sii in dvxc_asii.items():
            m1 = max(n1, gs.bd.beg) - gs.bd.beg
            m2 = min(n2, gs.bd.end) - gs.bd.beg
            if m1 < m2:
                P_ni = kpt.P_ani[a][m1:m2]
                vxc_n[m1:m2] += ((P_ni @ dvxc_sii[kpt.s]) * P_ni.conj()).sum(1).real

    gs.gd.comm.sum(vxc_un)
    vxc_skn = gs.kd.collect(vxc_un, broadcast=True)

    vxc_skn = gs.bd.collect(vxc_skn.T.copy(), broadcast=True).T
    return vxc_skn[:, :, n1:n2] * Ha


def vxc_mat(gs, xc=None, coredensity=True, n1=0, n2=0):
    """Calculate XC-contribution to Hamiltonian matrix elements in orbital space."""

    if n2 <= 0:
        n2 += gs.bd.nbands

    ham = gs.hamiltonian
    dens = gs.density

    if xc is None:
        xc = ham.xc
    elif isinstance(xc, str):
        xc = XC(xc)

    if dens.nt_sg is None:
        dens.interpolate_pseudo_density()

    xc.set_grid_descriptor(gs._density.finegd)
    xc.initialize(gs._density, gs._hamiltonian, gs._wfs)
    xc.set_positions(gs.spos_ac)
    gs._hamiltonian.get_xc_difference(xc, gs._density)

    # Calculate XC-potential
    vxct_sg = ham.finegd.zeros(gs.nspins)
    xc.calculate(dens.finegd, dens.nt_sg, vxct_sg)
    vxct_sG = ham.restrict_and_collect(vxct_sg)

    # Calculate PAW corrections:
    dvxc_asii = {}
    if isinstance(xc, CiderPASDW_MPRoutines):
        xc._collect_paw_corrections()
    for a, D_sp in dens.D_asp.items():
        dvxc_sp = np.zeros_like(D_sp)

        pawdata = gs.pawdatasets.by_atom[a]
        xc.calculate_paw_correction(
            pawdata, D_sp, dvxc_sp, a=a, addcoredensity=coredensity
        )
        dvxc_asii[a] = [unpack_hermitian(dvxc_p) for dvxc_p in dvxc_sp]

    vxc_n_skn = []
    for n0 in range(gs.bd.mynbands):
        if n1 <= n0 + gs.bd.beg < n2:
            vxc_un = np.empty(
                (gs.kd.mynk * gs.nspins, gs.bd.mynbands), dtype=np.complex128
            )
            for u, vxc_n in enumerate(vxc_un):
                kpt = gs.kpt_u[u]
                vxct_G = vxct_sG[kpt.s]
                if xc.type == "MGGA":
                    G_Gv = gs._wfs.pd.get_reciprocal_vectors(q=kpt.q)

                psit0_G = gs.get_wave_function_array(u, n0)
                tmp0_vG = [None] * 3
                if xc.type == "MGGA":
                    psit_k = kpt.psit_nG[n0]
                    for v in range(3):
                        tmp0_vG[v] = gs.pd.ifft(1j * G_Gv[:, v] * psit_k, q=kpt.q)

                for n in range(gs.bd.mynbands):
                    if n1 <= n + gs.bd.beg < n2:
                        psit_G = gs.get_wave_function_array(u, n)
                        vxc_n[n] = gs.gd.integrate(
                            (psit_G * psit0_G.conj()).real.copy(),
                            vxct_G,
                            global_integral=False,
                        ) + 1j * gs.gd.integrate(
                            (psit_G * psit0_G.conj()).imag.copy(),
                            vxct_G,
                            global_integral=False,
                        )
                        if xc.type == "MGGA":
                            psit_k = kpt.psit_nG[n]
                            dpsit2_G = np.zeros_like(vxct_G, dtype=complex)
                            for v in range(3):
                                tmp_G = gs.pd.ifft(1j * G_Gv[:, v] * psit_k, q=kpt.q)
                                dpsit2_G += 0.5 * (tmp_G * tmp0_vG[v].conj())
                            vxc_n[n] += gs.gd.integrate(
                                dpsit2_G.real.copy(),
                                xc.dedtaut_sG[kpt.s],
                                global_integral=False,
                            ) + 1j * gs.gd.integrate(
                                dpsit2_G.imag.copy(),
                                xc.dedtaut_sG[kpt.s],
                                global_integral=False,
                            )

                for a, dvxc_sii in dvxc_asii.items():
                    m1 = max(n1, gs.bd.beg) - gs.bd.beg
                    m2 = min(n2, gs.bd.end) - gs.bd.beg
                    if m1 < m2:
                        P_ni = kpt.P_ani[a][m1:m2]
                        vxc_n[m1:m2] += (
                            (P_ni @ dvxc_sii[kpt.s]) * P_ni[n0 - gs.bd.beg - m1].conj()
                        ).sum(1)

            gs.gd.comm.sum(vxc_un)
            vxc_skn = gs.kd.collect(vxc_un, broadcast=True)

            vxc_skn = gs.bd.collect(vxc_skn.T.copy(), broadcast=True).T
            vxc_n_skn.append(vxc_skn[:, :, n1:n2] * Ha)

    res = np.stack(vxc_n_skn, axis=-1)
    return res



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

from abc import ABC, abstractmethod

import numpy as np
import yaml


class FeatureNormalizer(ABC):
    code = None

    @property
    @abstractmethod
    def num_arg(self):
        pass

    @abstractmethod
    def bounds(self):
        pass

    @abstractmethod
    def fill_feat_(self, y, x):
        pass

    @abstractmethod
    def fill_deriv_(self, dfdx, dfdy, x):
        pass

    @classmethod
    def from_dict(cls, d):
        if d["code"] in ALL_CLASS_DICT:
            return ALL_CLASS_DICT[d["code"]].from_dict(d)
        else:
            raise ValueError("Unrecognized code: {}".format(d["code"]))


class LMap(FeatureNormalizer):
    code = "L"

    def __init__(self, i, bounds=None):
        self.i = i
        self._bounds = bounds or (-np.inf, np.inf)

    @property
    def bounds(self):
        return self._bounds

    @property
    def num_arg(self):
        return 1

    def fill_feat_(self, y, x):
        y[:] = x[self.i]

    def fill_deriv_(self, dfdx, dfdy, x):
        dfdx[self.i] += dfdy[:]

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "bounds": self.bounds,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], bounds=d.get("bounds"))


class UMap(FeatureNormalizer):
    code = "U"

    def __init__(self, i, gamma, bounds=None):
        self.i = i
        self.gamma = gamma
        self._bounds = bounds or (0, 1)

    @property
    def bounds(self):
        return self._bounds

    @property
    def num_arg(self):
        return 1

    def fill_feat_(self, y, x):
        y[:] = self.gamma * x[self.i] / (1 + self.gamma * x[self.i])

    def fill_deriv_(self, dfdx, dfdy, x):
        i = self.i
        dfdx[i] += dfdy * self.gamma / (1 + self.gamma * x[i]) ** 2

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "gamma": self.gamma,
            "bounds": self.bounds,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["gamma"], bounds=d.get("bounds"))


class TMap(FeatureNormalizer):
    code = "T"

    def __init__(self, i, j, bounds=None):
        self.i = i
        self.j = j
        self._bounds = bounds or (0, 1)

    @property
    def bounds(self):
        return self._bounds

    @property
    def num_arg(self):
        return 1

    def fill_feat_(self, y, x):
        i, j = self.i, self.j
        tmp = x[j] + 5.0 / 3 * x[i]
        y[:] = tmp / (1 + tmp)

    def fill_deriv_(self, dfdx, dfdy, x):
        i, j = self.i, self.j
        tmp = x[j] + 5.0 / 3 * x[i]
        dtmp = 1.0 / ((1 + tmp) * (1 + tmp))
        dtmp *= dfdy
        dfdx[i] += 5.0 / 3 * dtmp
        dfdx[j] += dtmp

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "j": self.j,
            "bounds": self.bounds,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["j"], bounds=d.get("bounds"))


def get_vmap_heg_value(heg, gamma):
    return (heg * gamma) / (1 + heg * gamma)


class VMap(FeatureNormalizer):
    code = "V"

    def __init__(self, i, gamma, scale=1.0, center=0.0, bounds=None):
        self.i = i
        self.gamma = gamma
        self.scale = scale
        self.center = center
        self._bounds = bounds or (-self.center, self.scale - self.center)

    @property
    def bounds(self):
        return self._bounds

    @property
    def num_arg(self):
        return 1

    def fill_feat_(self, y, x):
        i = self.i
        y[:] = -self.center + self.scale * self.gamma * x[i] / (1 + self.gamma * x[i])

    def fill_deriv_(self, dfdx, dfdy, x):
        i = self.i
        dfdx[i] += dfdy * self.scale * self.gamma / (1 + self.gamma * x[i]) ** 2

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "gamma": self.gamma,
            "scale": self.scale,
            "center": self.center,
            "bounds": self.bounds,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["gamma"], d["scale"], d["center"], bounds=d.get("bounds"))


class VZMap(FeatureNormalizer):
    code = "VZ"

    def __init__(self, i, gamma, scale=1.0, center=0.0, bounds=None):
        self.i = i
        self.gamma = gamma
        self.scale = scale
        self.center = center
        self._bounds = bounds or (-self.center, self.scale - self.center)

    @property
    def bounds(self):
        return self._bounds

    @property
    def num_arg(self):
        return 1

    def fill_feat_(self, y, x):
        xi = x[self.i]
        y[:] = -self.center + self.scale * self.gamma * (xi + xi * xi) / (
            1 + self.gamma * (xi + xi * xi)
        )

    def fill_deriv_(self, dfdx, dfdy, x):
        xi = x[self.i]
        fac = dfdy * self.scale * self.gamma
        fac /= (1 + self.gamma * (xi + xi * xi)) ** 2
        dfdx[self.i] += fac
        dfdx[self.i] += fac * (self.gamma + 1) * xi
        dfdx[self.i] += fac * (self.gamma - 1) * xi * (3 * xi + 2 * xi * xi)

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "gamma": self.gamma,
            "scale": self.scale,
            "center": self.center,
            "bounds": self.bounds,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["gamma"], d["scale"], d["center"], bounds=d.get("bounds"))


class V2Map(FeatureNormalizer):
    code = "V2"

    def __init__(self, i, j):
        self.i = i
        self.j = j

    @property
    def bounds(self):
        return (-1.0, 1.0)

    @property
    def num_arg(self):
        return 1

    def fill_feat_(self, y, x):
        i, j = self.i, self.j
        a = 2**1.5
        b = 0.5 / (a - 1)
        tmp = b * (a * x[i] - x[j])
        y[:] = tmp / (1 + tmp)
        tmp = 0.5 * x[j]
        y[:] -= tmp / (1 + tmp)

    def fill_deriv_(self, dfdx, dfdy, x):
        i, j = self.i, self.j
        a = 2**1.5
        b = 0.5 / (a - 1)
        tmp = a * x[i] - x[j]
        tmp = b / (1 + b * tmp) ** 2
        dfdx[i] += dfdy * tmp * a
        dfdx[j] -= dfdy * tmp
        tmp = 0.5 / (1 + 0.5 * x[j]) ** 2
        dfdx[j] -= dfdy * tmp

    def as_dict(self):
        return {"code": self.code, "i": self.i, "j": self.j}

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["j"])


class V3Map(FeatureNormalizer):
    code = "V3"

    def __init__(self, i, j, gamma):
        self.i = i
        self.j = j
        self.gamma = gamma

    @property
    def bounds(self):
        return (-1.0, 1.0)

    @property
    def num_arg(self):
        return 2

    def fill_feat_(self, y, x):
        i, j = self.i, self.j
        y[:] = self.gamma * x[i] / (1 + self.gamma * x[i])
        y[:] -= self.gamma * x[j] / (1 + self.gamma * x[j])

    def fill_deriv_(self, dfdx, dfdy, x):
        i, j = self.i, self.j
        dfdx[i] += dfdy * self.gamma / (1 + self.gamma * x[i]) ** 2
        dfdx[j] -= dfdy * self.gamma / (1 + self.gamma * x[j]) ** 2

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "j": self.j,
            "gamma": self.gamma,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["j"], d["gamma"])


class V4Map(FeatureNormalizer):
    code = "V4"

    def __init__(self, i, j, gamma):
        self.i = i
        self.j = j
        self.gamma = gamma

    @property
    def bounds(self):
        return (-0.5, 0.5)

    @property
    def num_arg(self):
        return 2

    def fill_feat_(self, y, x):
        i, j = self.i, self.j
        tmp = np.exp(self.gamma * (x[i] - x[j]))
        y[:] = 1 / (1 + tmp) - 0.5

    def fill_deriv_(self, dfdx, dfdy, x):
        i, j = self.i, self.j
        tmp = np.exp(self.gamma * (x[i] - x[j]))
        tmp = dfdy * self.gamma * tmp / (1 + tmp) ** 2
        dfdx[i] -= tmp
        dfdx[j] += tmp

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "j": self.j,
            "gamma": self.gamma,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["j"], d["gamma"])


class WMap(FeatureNormalizer):
    code = "W"

    def __init__(self, i, j, k, gammai, gammaj):
        self.i = i
        self.j = j
        self.k = k
        self.gammai = gammai
        self.gammaj = gammaj

    @property
    def bounds(self):
        return (-1, 1)

    @property
    def num_arg(self):
        return 3

    def fill_feat_(self, y, x):
        i, j, k = self.i, self.j, self.k
        gammai, gammaj = self.gammai, self.gammaj
        y[:] = (gammai * np.sqrt(gammaj / (1 + gammaj * x[j])) * x[k]) / (
            1 + gammai * x[i]
        )

    def fill_deriv_(self, dfdx, dfdy, x):
        i, j, k = self.i, self.j, self.k
        gammai, gammaj = self.gammai, self.gammaj
        dfdx[i] -= dfdy * (
            (gammai**2 * np.sqrt(gammaj / (1 + gammaj * x[j])) * x[k])
            / (1 + gammai * x[i]) ** 2
        )
        dfdx[j] -= (
            dfdy
            * (gammai * (gammaj / (1 + gammaj * x[j])) ** 1.5 * x[k])
            / (2.0 * (1 + gammai * x[i]))
        )
        dfdx[k] += (
            dfdy
            * (gammai * np.sqrt(gammaj / (1 + gammaj * x[j])))
            / (1 + gammai * x[i])
        )

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "j": self.j,
            "k": self.k,
            "gammai": self.gammai,
            "gammaj": self.gammaj,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["j"], d["k"], d["gammai"], d["gammaj"])


class XMap(FeatureNormalizer):
    code = "X"

    def __init__(self, i, j, k, gammai, gammaj):
        self.i = i
        self.j = j
        self.k = k
        self.gammai = gammai
        self.gammaj = gammaj

    @property
    def bounds(self):
        return (-1, 1)

    @property
    def num_arg(self):
        return 3

    def fill_feat_(self, y, x):
        i, j, k = self.i, self.j, self.k
        gammai, gammaj = self.gammai, self.gammaj
        y[:] = (
            np.sqrt(gammai / (1 + gammai * x[i]))
            * np.sqrt(gammaj / (1 + gammaj * x[j]))
            * x[k]
        )

    def fill_deriv_(self, dfdx, dfdy, x):
        i, j, k = self.i, self.j, self.k
        gammai, gammaj = self.gammai, self.gammaj
        dfdx[i] -= dfdy * (
            (
                gammai
                * np.sqrt(gammai / (1 + gammai * x[i]))
                * np.sqrt(gammaj / (1 + gammaj * x[j]))
                * x[k]
            )
            / (2 + 2 * gammai * x[i])
        )
        dfdx[j] -= dfdy * (
            (
                gammaj
                * np.sqrt(gammai / (1 + gammai * x[i]))
                * np.sqrt(gammaj / (1 + gammaj * x[j]))
                * x[k]
            )
            / (2 + 2 * gammaj * x[j])
        )
        dfdx[k] += (
            dfdy
            * np.sqrt(gammai / (1 + gammai * x[i]))
            * np.sqrt(gammaj / (1 + gammaj * x[j]))
        )

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "j": self.j,
            "k": self.k,
            "gammai": self.gammai,
            "gammaj": self.gammaj,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["j"], d["k"], d["gammai"], d["gammaj"])


class YMap(FeatureNormalizer):
    code = "Y"

    def __init__(self, i, j, k, l, gammai, gammaj, gammak):
        self.i = i
        self.j = j
        self.k = k
        self.l = l
        self.gammai = gammai
        self.gammaj = gammaj
        self.gammak = gammak

    @property
    def bounds(self):
        return (-1, 1)

    @property
    def num_arg(self):
        return 4

    def fill_feat_(self, y, x):
        i, j, k, l = self.i, self.j, self.k, self.l
        gammai, gammaj, gammak = self.gammai, self.gammaj, self.gammak
        y[:] = (
            np.sqrt(gammai / (1 + gammai * x[i]))
            * x[l]
            * np.sqrt(gammaj / (1 + gammaj * x[j]))
            * np.sqrt(gammak / (1 + gammak * x[k]))
        )

    def fill_deriv_(self, dfdx, dfdy, x):
        i, j, k, l = self.i, self.j, self.k, self.l
        gammai, gammaj, gammak = self.gammai, self.gammaj, self.gammak
        dfdx[i] -= dfdy * (
            (
                gammai
                * np.sqrt(gammai / (1 + gammai * x[i]))
                * x[l]
                * np.sqrt(gammaj / (1 + gammaj * x[j]))
                * np.sqrt(gammak / (1 + gammak * x[k]))
            )
            / (2 + 2 * gammai * x[i])
        )
        dfdx[j] -= dfdy * (
            (
                gammaj
                * np.sqrt(gammai / (1 + gammai * x[i]))
                * x[l]
                * np.sqrt(gammaj / (1 + gammaj * x[j]))
                * np.sqrt(gammak / (1 + gammak * x[k]))
            )
            / (2 + 2 * gammaj * x[j])
        )
        dfdx[k] -= dfdy * (
            (
                gammak
                * np.sqrt(gammai / (1 + gammai * x[i]))
                * x[l]
                * np.sqrt(gammaj / (1 + gammaj * x[j]))
                * np.sqrt(gammak / (1 + gammak * x[k]))
            )
            / (2 + 2 * gammak * x[k])
        )
        dfdx[l] += (
            dfdy
            * np.sqrt(gammai / (1 + gammai * x[i]))
            * np.sqrt(gammaj / (1 + gammaj * x[j]))
            * np.sqrt(gammak / (1 + gammak * x[k]))
        )

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "j": self.j,
            "k": self.k,
            "l": self.l,
            "gammai": self.gammai,
            "gammaj": self.gammaj,
            "gammak": self.gammak,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(
            d["i"], d["j"], d["k"], d["l"], d["gammai"], d["gammaj"], d["gammak"]
        )


class ZMap(FeatureNormalizer):
    code = "Z"

    def __init__(self, i, gamma, scale=1.0, center=0.0, bounds=None):
        self.i = i
        self.gamma = gamma
        self.scale = scale
        self.center = center
        self._bounds = bounds or (-self.center, self.scale - self.center)

    @property
    def bounds(self):
        return self._bounds

    @property
    def num_arg(self):
        return 1

    def fill_feat_(self, y, x):
        i = self.i
        y[:] = -self.center + self.scale / (1 + self.gamma * x[i] ** 2)

    def fill_deriv_(self, dfdx, dfdy, x):
        i = self.i
        dfdx[i] -= (
            2
            * dfdy
            * self.scale
            * self.gamma
            * x[i]
            / (1 + self.gamma * x[i] ** 2) ** 2
        )

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "gamma": self.gamma,
            "scale": self.scale,
            "center": self.center,
            "bounds": self.bounds,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["gamma"], d["scale"], d["center"], bounds=d.get("bounds"))


class EMap(FeatureNormalizer):
    code = "E"

    def __init__(self, i, scale=1.0, center=0.0, bounds=None):
        self.i = i
        self.scale = scale
        self.center = center
        self._bounds = bounds or (0, 1)

    @property
    def bounds(self):
        return self._bounds

    @property
    def num_arg(self):
        return 1

    def fill_feat_(self, y, x):
        i = self.i
        y[:] = np.exp(-self.scale * x[i] + self.center)

    def fill_deriv_(self, dfdx, dfdy, x):
        i = self.i
        dfdx[i] -= dfdy * self.scale * np.exp(-self.scale * x[i] + self.center)

    def as_dict(self):
        return {
            "code": self.code,
            "i": self.i,
            "scale": self.scale,
            "center": self.center,
            "bounds": self.bounds,
        }

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["scale"], d["center"], bounds=d.get("bounds"))


class SignedUMap(FeatureNormalizer):
    code = "SU"

    def __init__(self, i, gamma):
        self.i = i
        self.gamma = gamma

    @property
    def bounds(self):
        return (-1, 1)

    @property
    def num_arg(self):
        return 1

    def fill_feat_(self, y, x):
        i = self.i
        y[:] = x[i] / np.sqrt(self.gamma + x[i] * x[i])

    def fill_deriv_(self, dfdx, dfdy, x):
        i = self.i
        dfdx[i] += dfdy * self.gamma / (self.gamma + x[i] * x[i]) ** 1.5

    def as_dict(self):
        return {"code": self.code, "i": self.i, "gamma": self.gamma}

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["gamma"])


class SLNMap(FeatureNormalizer):
    code = "SLN"

    def __init__(self, i, gamma):
        self.i = i
        self.gamma = gamma

    @property
    def bounds(self):
        return (0, 1)

    @property
    def num_arg(self):
        return 1

    def fill_feat_(self, y, x):
        rho = x[self.i]
        const = 3.0 * self.gamma / (4 * np.pi)
        rss = (1 + const * rho) ** (-1.0 / 3)
        y[:] = rss

    def fill_deriv_(self, dfdx, dfdy, x):
        rho = x[self.i]
        const = 3.0 * self.gamma / (4 * np.pi)
        dfdx[self.i] -= dfdy * const / 3.0 * (1 + const * rho) ** (-4.0 / 3)

    def as_dict(self):
        return {"code": self.code, "i": self.i, "gamma": self.gamma}

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["gamma"])


class SLXMap(FeatureNormalizer):
    """
    gamma * s^2 / (1 + gamma * s^2), where s^2 is the squared reduced gradient
    s^2 = sigma / C n^8/3
    """

    code = "SLX"

    def __init__(self, i, j, gamma):
        self.i = i
        self.j = j
        self.gamma = gamma

    @property
    def bounds(self):
        return (0, 1)

    @property
    def num_arg(self):
        return 2

    def fill_feat_(self, y, x):
        rho = np.maximum(x[self.i], 1e-10)
        rho83 = rho ** (8.0 / 3)
        const = 2 * (3 * np.pi**2) ** (1.0 / 3)
        const = self.gamma / const**2
        sigma = const * x[self.j]
        y[:] = sigma / (rho83 + sigma)

    def fill_deriv_(self, dfdx, dfdy, x):
        rho = np.maximum(x[self.i], 1e-10)
        rho83 = rho ** (8.0 / 3)
        rho53 = (8.0 / 3) * rho ** (5.0 / 3)
        const = 2 * (3 * np.pi**2) ** (1.0 / 3)
        const = self.gamma / const**2
        sigma = const * x[self.j]
        fac = rho83 + sigma
        fac[:] = fac * fac
        fac[:] = dfdy / fac
        dfdx[self.i] -= fac * sigma * rho53
        dfdx[self.j] += fac * rho83 * const

    def as_dict(self):
        return {"code": self.code, "i": self.i, "j": self.j, "gamma": self.gamma}

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["j"], d["gamma"])


class SLBMap(FeatureNormalizer):
    """
    beta = (tau - tauw) / (tau + tau0)
    """

    code = "SLB"

    const = 0.3 * (3 * np.pi**2) ** (2.0 / 3)

    def __init__(self, i, j, k):
        self.i = i
        self.j = j
        self.k = k

    @property
    def bounds(self):
        return (-1, 1)

    @property
    def num_arg(self):
        return 3

    def fill_feat_(self, y, x):
        rho = np.maximum(x[self.i], 1e-10)
        tau0 = self.const * rho ** (5.0 / 3)
        tauw = x[self.j] / (8 * rho)
        tau = x[self.k]
        # y[:] = -1 + 2 * (tau - tauw) / (tau0 + tau - tauw)
        y[:] = -1 + 2 * (tau - tauw) / (tau + tau0)

    def fill_deriv_(self, dfdx, dfdy, x):
        rho = np.maximum(x[self.i], 1e-10)
        tau0 = self.const * rho ** (5.0 / 3)
        tauw = x[self.j] / (8 * rho)
        tau = x[self.k]
        fac = 1.0 / (tau + tau0)
        v0 = 2 * dfdy * (tauw - tau) * fac * fac
        vw = -2 * dfdy * fac
        vt = 2 * dfdy * (tau0 + tauw) * fac * fac
        dfdx[self.i] += v0 * self.const * (5.0 / 3) * rho ** (2.0 / 3)
        dfdx[self.i] -= vw * tauw / rho
        dfdx[self.j] += vw / (8 * rho)
        dfdx[self.k] += vt

    def as_dict(self):
        return {"code": self.code, "i": self.i, "j": self.j, "k": self.k}

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["j"], d["k"])


class SLTMap(FeatureNormalizer):
    """
    beta = (tau - tau0) / (tau + tau0)
    """

    code = "SLT"

    const = 0.3 * (3 * np.pi**2) ** (2.0 / 3)

    def __init__(self, i, j):
        self.i = i
        self.j = j

    @property
    def bounds(self):
        return (-1, 1)

    @property
    def num_arg(self):
        return 2

    def fill_feat_(self, y, x):
        rho = np.maximum(x[self.i], 1e-10)
        tau0 = self.const * rho ** (5.0 / 3)
        tau = x[self.j]
        y[:] = (tau - tau0) / (tau + tau0)

    def fill_deriv_(self, dfdx, dfdy, x):
        rho = np.maximum(x[self.i], 1e-10)
        tau0 = self.const * rho ** (5.0 / 3)
        tau = x[self.j]
        fac = 1.0 / (tau + tau0)
        v0 = -2 * dfdy * tau * fac * fac
        vt = 2 * dfdy * tau0 * fac * fac
        dfdx[self.i] += v0 * self.const * (5.0 / 3) * rho ** (2.0 / 3)
        dfdx[self.j] += vt

    def as_dict(self):
        return {"code": self.code, "i": self.i, "j": self.j}

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["j"])


class SLTWMap(FeatureNormalizer):
    """
    beta = (tauw - tau0) / (tauw + tau0)
    """

    code = "SLTW"

    const = 0.3 * (3 * np.pi**2) ** (2.0 / 3)

    def __init__(self, i, j):
        self.i = i
        self.j = j

    @property
    def bounds(self):
        return (-1, 1)

    @property
    def num_arg(self):
        return 2

    def fill_feat_(self, y, x):
        rho = np.maximum(x[self.i], 1e-10)
        tau0 = self.const * rho ** (5.0 / 3)
        tauw = x[self.j] / (8 * rho)
        y[:] = (tauw - tau0) / (tauw + tau0)

    def fill_deriv_(self, dfdx, dfdy, x):
        rho = np.maximum(x[self.i], 1e-10)
        tau0 = self.const * rho ** (5.0 / 3)
        tauw = x[self.j] / (8 * rho)
        fac = 1.0 / (tauw + tau0)
        v0 = -2 * dfdy * tauw * fac * fac
        vw = 2 * dfdy * tau0 * fac * fac
        dfdx[self.i] += v0 * self.const * (5.0 / 3) * rho ** (2.0 / 3)
        dfdx[self.i] -= vw * tauw / rho
        dfdx[self.j] += vw / (8 * rho)

    def as_dict(self):
        return {"code": self.code, "i": self.i, "j": self.j}

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["j"])


class SLDMap(FeatureNormalizer):
    code = "SLD"
    const = 0.3 * (3 * np.pi**2) ** (2.0 / 3)

    def __init__(self, i, j, k):
        self.i = i
        self.j = j
        self.k = k

    @property
    def bounds(self):
        return (-1, 1)

    @property
    def num_arg(self):
        return 3

    def fill_feat_(self, y, x):
        rho = np.maximum(x[self.i], 1e-10)
        tau0 = self.const * rho ** (5.0 / 3)
        tauw = x[self.j] / (8 * rho)
        tau = x[self.k]
        y[:] = tau / (tau + tau0)
        y[:] -= tauw / (tauw + tau0)

    def fill_deriv_(self, dfdx, dfdy, x):
        rho = np.maximum(x[self.i], 1e-10)
        tau0 = self.const * rho ** (5.0 / 3)
        tauw = x[self.j] / (8 * rho)
        tau = x[self.k]
        fac1 = 1.0 / (tau + tau0)
        fac2 = 1.0 / (tauw + tau0)
        v0 = -dfdy * tau * fac1 * fac1
        v0 += dfdy * tauw * fac2 * fac2
        vw = -dfdy * tau0 * fac2 * fac2
        vt = dfdy * tau0 * fac1 * fac1
        dfdx[self.i] += v0 * self.const * (5.0 / 3) * rho ** (2.0 / 3)
        dfdx[self.i] -= vw * tauw / rho
        dfdx[self.j] += vw / (8 * rho)
        dfdx[self.k] += vt

    def as_dict(self):
        return {"code": self.code, "i": self.i, "j": self.j, "k": self.k}

    @classmethod
    def from_dict(cls, d):
        return cls(d["i"], d["j"], d["k"])


ALL_CLASS_DICT = {}
ALL_CLASSES = [
    LMap,
    UMap,
    TMap,
    VMap,
    VZMap,
    V2Map,
    V3Map,
    V4Map,
    WMap,
    XMap,
    YMap,
    ZMap,
    EMap,
    SignedUMap,
    SLNMap,
    SLXMap,
    SLBMap,
    SLTMap,
    SLTWMap,
    SLDMap,
]
for cls in ALL_CLASSES:
    assert cls.code not in ALL_CLASS_DICT
    ALL_CLASS_DICT[cls.code] = cls


class FeatureList:
    def __init__(self, feat_list):
        self.feat_list = feat_list

    @property
    def nfeat(self):
        return len(self.feat_list)

    def __getitem__(self, index):
        return self.feat_list[index]

    def __call__(self, xdesc):
        # xdesc (nsamp, ninp)
        xdesc = xdesc.T
        # now xdesc (ninp, nsamp)
        tdesc = np.zeros((self.nfeat, xdesc.shape[1]))
        for i in range(self.nfeat):
            self.feat_list[i].fill_feat_(tdesc[i], xdesc)
        return tdesc.T

    @property
    def bounds_list(self):
        return [f.bounds for f in self.feat_list]

    def fill_vals_(self, tdesc, xdesc):
        for i in range(self.nfeat):
            self.feat_list[i].fill_feat_(tdesc[i], xdesc)

    def fill_derivs_(self, dfdx, dfdy, xdesc):
        """
        Fill dfdx wth the derivatives of f with respect
        to the raw features x, given the derivative
        with respect to transformed features dfdy.

        Args:
            dfdx (nraw, nsamp): Output. Derivative of output wrt
                raw features x
            dfdy (ntrans, nsamp): Input. Derivative of output wrt transformed
                features y
            xdesc (nraw, nsamp): Raw features x
        """
        for i in range(self.nfeat):
            self.feat_list[i].fill_deriv_(dfdx, dfdy[i], xdesc)

    def as_dict(self):
        d = {"feat_list": [f.as_dict() for f in self.feat_list]}
        return d

    def dump(self, fname):
        d = self.as_dict()
        with open(fname, "w") as f:
            yaml.dump(d, f)

    @classmethod
    def from_dict(cls, d):
        return cls(
            [
                FeatureNormalizer.from_dict(d["feat_list"][i])
                for i in range(len(d["feat_list"]))
            ]
        )

    @classmethod
    def load(cls, fname):
        with open(fname, "r") as f:
            d = yaml.load(f, Loader=yaml.Loader)
        return cls.from_dict(d)


"""
Examples of FeatureList objects:
center0a = get_vmap_heg_value(2.0, gamma0a)
center0b = get_vmap_heg_value(8.0, gamma0b)
center0c = get_vmap_heg_value(0.5, gamma0c)
lst = [
        UMap(1, gammax),
        VMap(2, 1, scale=2.0, center=1.0),
        VMap(3, gamma0a, scale=1.0, center=center0a),
        UMap(4, gamma1),
        UMap(5, gamma2),
        WMap(1, 5, 6, gammax, gamma2),
        XMap(1, 4, 7, gammax, gamma1),
        VMap(8, gamma0b, scale=1.0, center=center0b),
        VMap(9, gamma0c, scale=1.0, center=center0c),
        YMap(1, 4, 5, 10, gammax, gamma1, gamma2),
]
flst = FeatureList(lst)

center0a = get_vmap_heg_value(2.0, gamma0a)
center0b = get_vmap_heg_value(3.0, gamma0b)
UMap(1, gammax)
VMap(2, 1, scale=2.0, center=1.0)
VMap(3, gamma0a, scale=1.0, center=center0a)
UMap(4, gamma1)
UMap(5, gamma2)
WMap(1, 5, 6, gammax, gamma2)
XMap(1, 4, 7, gammax, gamma1)
VMap(8, gamma0b, scale=1.0, center=center0b)
YMap(1, 4, 5, 9, gammax, gamma1, gamma2)
"""



import joblib
import numpy as np
import yaml

from ciderpress.gpaw.cider_kernel import CiderGGAHybridKernel, CiderMGGAHybridKernel
from ciderpress.gpaw.interp_paw import DiffGGA, DiffMGGA


class SLCiderGGAHybridWrapper(CiderGGAHybridKernel):
    def calculate(self, e_g, n_sg, v_sg, sigma_xg, dedsigma_xg):
        super(SLCiderGGAHybridWrapper, self).calculate(
            e_g,
            n_sg,
            v_sg,
            sigma_xg,
            dedsigma_xg,
            np.zeros(
                (
                    n_sg.shape[0],
                    3,
                )
                + e_g.shape,
                dtype=e_g.dtype,
            ),
        )


class SLCiderMGGAHybridWrapper(CiderMGGAHybridKernel):
    def calculate(self, e_g, n_sg, v_sg, sigma_xg, dedsigma_xg, tau_sg, dedtau_sg):
        super(SLCiderMGGAHybridWrapper, self).calculate(
            e_g,
            n_sg,
            v_sg,
            sigma_xg,
            dedsigma_xg,
            tau_sg,
            dedtau_sg,
            np.zeros(
                (
                    n_sg.shape[0],
                    3,
                )
                + e_g.shape,
                dtype=e_g.dtype,
            ),
        )


class _SLCiderBase:
    def get_setup_name(self):
        return "PBE"

    def todict(self):
        kparams = self.kernel.todict()
        return {
            "kernel_params": kparams,
        }

    def get_mlfunc_data(self):
        return yaml.dump(self.kernel.mlfunc, Dumper=yaml.CDumper)

    @classmethod
    def from_joblib(cls, fname, **kwargs):
        mlfunc = joblib.load(fname)
        return cls.from_mlfunc(mlfunc, **kwargs)

    @staticmethod
    def from_mlfunc(mlfunc, xmix=1.00, xkernel="GGA_X_PBE", ckernel="GGA_C_PBE"):
        if mlfunc.desc_version == "b":
            cider_kernel = SLCiderMGGAHybridWrapper(mlfunc, xmix, xkernel, ckernel)
            cls = SLCiderMGGA
        elif mlfunc.desc_version == "d":
            cider_kernel = SLCiderGGAHybridWrapper(mlfunc, xmix, xkernel, ckernel)
            cls = SLCiderGGA
        else:
            raise ValueError(
                "Only implemented for b and d version, found {}".format(
                    mlfunc.desc_version
                )
            )

        return cls(cider_kernel)


class SLCiderGGA(_SLCiderBase, DiffGGA):
    def todict(self):
        d = super(SLCiderGGA, self).todict()
        d["_cider_type"] = "SLCiderGGA"
        return d


class SLCiderMGGA(_SLCiderBase, DiffMGGA):
    def todict(self):
        d = super(SLCiderMGGA, self).todict()
        d["_cider_type"] = "SLCiderMGGA"
        return d



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

from itertools import combinations

import numpy as np
import pyscf.lib
from interpolation.splines import UCGrid, filter_cubic

from ciderpress.models.kernels import (
    DiffAdditiveMixin,
    DiffARBF,
    DiffProduct,
    SubsetAddLLRBF,
    SubsetAddRQ,
    SubsetARBF,
    SubsetRBF,
    arbf_args,
)

"""
Script for mapping a CIDER GP to a cubic spline.
Requires an input DFTGPR object, stored in joblib format.
"""


def get_dim(x, length_scale, density=6, buff=0.0, bound=None, max_ngrid=None):
    print(length_scale, bound)
    mini = np.min(x) - buff
    maxi = np.max(x) + buff
    if bound is not None:
        mini, maxi = bound[0], bound[1]
    ran = maxi - mini
    ngrid = max(int(density * ran / length_scale) + 1, 3)
    if max_ngrid is not None and ngrid > max_ngrid:
        ngrid = max_ngrid
    return (mini, maxi, ngrid)


def project_kernel_onto_grid(alpha, k0s, dims):
    N = len(dims)
    dims = [(float(dim[0]), float(dim[1]), int(dim[2])) for dim in dims]
    if N == 0:
        raise ValueError("Kernel is constant!")
    elif N == 1:
        fps = np.dot(alpha, k0s[0])
        sgd = UCGrid(dims[0])
    elif N == 2:
        k = np.einsum("ni,nj->nij", k0s[0], k0s[1])
        sgd = UCGrid(dims[0], dims[1])
        fps = np.einsum("n,nij->ij", alpha, k)
    elif N == 3:
        fps = 0
        for p0, p1 in pyscf.lib.prange(0, len(k0s[0]), 200):
            k = np.einsum("ni,nj,nk->nijk", k0s[0][p0:p1], k0s[1][p0:p1], k0s[2][p0:p1])
            fps += np.einsum("n,nijk->ijk", alpha[p0:p1], k)
        sgd = UCGrid(dims[0], dims[1], dims[2])
    elif N == 4:
        fps = 0
        for p0, p1 in pyscf.lib.prange(0, len(k0s[0]), 20):
            k = np.einsum(
                "ni,nj,nk,nl->nijkl",
                k0s[0][p0:p1],
                k0s[1][p0:p1],
                k0s[2][p0:p1],
                k0s[3][p0:p1],
            )
            fps += np.einsum("n,nijkl->ijkl", alpha[p0:p1], k)
        # k = np.einsum('ni,nj,nk,nl->nijkl', k0s[0], k0s[1], k0s[2], k0s[3])
        # funcps.append(np.einsum('n,nijkl->ijkl', alpha, k))
        sgd = UCGrid(dims[0], dims[1], dims[2], dims[3])
    else:
        raise ValueError("Order too high!")
    return fps, sgd


def get_mapped_gp_evaluator_linear(kernel, X, alpha):
    N = X.shape[1]
    assert N == alpha.size
    XT = np.identity(N)
    coefs = kernel(XT, X).dot(alpha)
    from ciderpress.dft.xc_evaluator import GlobalLinearEvaluator

    return GlobalLinearEvaluator(coefs)


def get_mapped_gp_evaluator_simple(
    rbf,
    X,
    alpha,
    feature_list,
    rbf_density=8,
    max_ngrid=120,
):
    """
    Quick approach for pure RBF GP mapping for N <= 4
    """
    N = X.shape[1]
    length_scale = rbf.k2.length_scale
    scale = [rbf.k1.constant_value]

    inds = np.arange(N)
    if isinstance(rbf.k2, SubsetRBF):
        inds = inds[rbf.k2.indexes]
    inds = list(inds)
    D = X[:, inds]
    N = D.shape[1]

    dims = []
    for i in range(N):
        dims.append(
            get_dim(
                D[:, i],
                length_scale[i],
                density=rbf_density,
                bound=feature_list[inds[i]].bounds,
                max_ngrid=max_ngrid,
            )
        )
    grid = [np.linspace(dims[i][0], dims[i][1], dims[i][2]) for i in range(N)]

    k0s = []
    for i in range(D.shape[1]):
        diff = (D[:, i : i + 1] - grid[i][np.newaxis, :]) / length_scale[i]
        k0s.append(np.exp(-0.5 * diff**2))

    funcp, spline_gd = project_kernel_onto_grid(alpha, k0s, dims)
    ind_sets = [inds]
    funcps = [funcp]
    spline_grids = [spline_gd]
    coeff_sets = [filter_cubic(spline_grids[0], funcps[0])]
    return scale, ind_sets, spline_grids, coeff_sets


def get_mapped_gp_evaluator_additive(
    kernel,
    X,
    alpha,
    feature_list,
    srbf_density=8,
    arbf_density=8,
    max_ngrid=120,
):
    N = X.shape[1]
    dims = []
    if isinstance(kernel, (DiffARBF, DiffAdditiveMixin)):
        srbf = None
        arbf = kernel
        ndim, length_scale, scale, order = arbf_args(arbf)
        sinds = np.zeros((0,))
        ainds = np.arange(N)[arbf.indexes]
        inds = ainds.copy()
    else:
        assert isinstance(kernel, DiffProduct)
        srbf = kernel.k1
        arbf = kernel.k2
        assert isinstance(srbf, SubsetRBF)
        assert isinstance(arbf, (SubsetARBF, SubsetAddRQ, SubsetAddLLRBF))
        ndim, length_scale, scale, order = arbf_args(arbf)
        length_scale = np.append(srbf.length_scale, length_scale)
        sinds = np.arange(N)[srbf.indexes]
        ainds = np.arange(N)[arbf.indexes]
        inds = np.append(sinds, ainds)

    D = X[:, inds]
    N = D.shape[1]
    for i in range(N):
        density = srbf_density if i < len(sinds) else arbf_density
        dims.append(
            get_dim(
                D[:, i],
                length_scale[i],
                density=density,
                bound=feature_list[inds[i]].bounds,
                max_ngrid=max_ngrid,
            )
        )
    grid = [np.linspace(dims[i][0], dims[i][1], dims[i][2]) for i in range(N)]
    k0s = []
    if isinstance(arbf, DiffARBF):
        for i in range(D.shape[1]):
            diff = (D[:, i : i + 1] - grid[i][np.newaxis, :]) / length_scale[i]
            k0s.append(np.exp(-0.5 * diff**2))
    elif isinstance(arbf, DiffAdditiveMixin):
        assert srbf is None
        for i in range(D.shape[1]):
            k0s.append(arbf.get_k0_for_mapping(D[:, i], grid[i], length_scale[i]))
    else:
        raise ValueError

    funcps = []
    spline_grids = []
    ind_sets = []
    arange_s = np.arange(len(sinds)).astype(int)
    arange_a = len(sinds) + np.arange(len(ainds)).astype(int)
    const = 0
    for o in range(order + 1):
        for tmp_inds in combinations(arange_a, o):
            if len(sinds > 0):
                tmp_inds = np.append(arange_s, tmp_inds).astype(int).tolist()
            else:
                tmp_inds = np.array(tmp_inds).astype(int).tolist()
            if len(tmp_inds) == 0:
                # raise NotImplementedError
                const += scale[0] * np.sum(alpha)
            else:
                ind_sets.append(list(inds[tmp_inds]))
                funcp, spline_gd = project_kernel_onto_grid(
                    alpha,
                    [k0s[i] for i in tmp_inds],
                    [dims[i] for i in tmp_inds],
                )
                funcps.append(funcp)
                spline_grids.append(spline_gd)

    coeff_sets = []
    for i in range(len(funcps)):
        coeff_sets.append(filter_cubic(spline_grids[i], funcps[i]))

    if len(sinds) == 0:
        scale = scale[1:]
        return scale, ind_sets, spline_grids, coeff_sets, const
    return scale, ind_sets, spline_grids, coeff_sets



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes

import numpy as np
from sympy.physics.wigner import clebsch_gordan

from ciderpress.lib import load_library as load_cider_library

libcider = load_cider_library("libmcider")


def change_basis_real_to_complex(l: int) -> np.ndarray:
    # https://en.wikipedia.org/wiki/Spherical_harmonics#Real_form
    q = np.zeros((2 * l + 1, 2 * l + 1), dtype=np.complex128)
    for m in range(-l, 0):
        q[l + m, l + abs(m)] = 1 / np.sqrt(2)
        q[l + m, l - abs(m)] = -1j / np.sqrt(2)
    q[l, l] = 1
    for m in range(1, l + 1):
        q[l + m, l + abs(m)] = (-1) ** m / np.sqrt(2)
        q[l + m, l - abs(m)] = 1j * (-1) ** m / np.sqrt(2)
    return (
        -1j
    ) ** l * q  # Added factor of 1j**l to make the Clebsch-Gordan coefficients real


def su2_clebsch_gordan(l1, l2, l3):
    mat = np.zeros(
        [
            2 * l1 + 1,
            2 * l2 + 1,
            2 * l3 + 1,
        ]
    )
    for m1 in range(2 * l1 + 1):
        for m2 in range(2 * l2 + 1):
            for m3 in range(2 * l3 + 1):
                mat[m1, m2, m3] = clebsch_gordan(l1, l2, l3, m1 - l1, m2 - l2, m3 - l3)
    return mat


def clebsch_gordan_e3nn(l1: int, l2: int, l3: int) -> np.ndarray:
    r"""
    The Clebsch-Gordan coefficients of the real irreducible representations of :math:`SO(3)`.
    This function taken from https://github.com/e3nn/e3nn

    Args:
        l1 (int): the representation order of the first irrep
        l2 (int): the representation order of the second irrep
        l3 (int): the representation order of the third irrep

    Returns:
        np.ndarray: the Clebsch-Gordan coefficients
    """
    C = su2_clebsch_gordan(l1, l2, l3)
    Q1 = change_basis_real_to_complex(l1)
    Q2 = change_basis_real_to_complex(l2)
    Q3 = change_basis_real_to_complex(l3)
    C = np.einsum("ij,kl,mn,ikn->jlm", Q1, Q2, np.conj(Q3.T), C)

    assert np.all(np.abs(np.imag(C)) < 1e-5)
    return np.real(C)


def get_ylm(r, lmax):
    N = r.shape[0]
    nlm = (lmax + 1) * (lmax + 1)
    rnorm = np.linalg.norm(r, axis=1)
    rhat = r / rnorm[:, np.newaxis]
    ylm = np.zeros((N, nlm))
    libcider.recursive_sph_harm_vec(
        ctypes.c_int(nlm),
        ctypes.c_int(N),
        rhat.ctypes.data_as(ctypes.c_void_p),
        ylm.ctypes.data_as(ctypes.c_void_p),
    )
    xyz = ylm[:, 1:4].copy()
    ylm[:, 3] = xyz[:, 0]
    ylm[:, 1] = xyz[:, 1]
    ylm[:, 2] = xyz[:, 2]
    for l in range(lmax + 1):
        for m in range(2 * l + 1):
            lm = l * l + m
            ylm[:, lm] *= rnorm**l
    return ylm


def get_ylm_fd(i, r, lmax, delta=1e-6):
    rx = r.copy()
    rx[:, i] += 0.5 * delta
    ylmp = get_ylm(rx, lmax)
    rx[:, i] -= delta
    ylmm = get_ylm(rx, lmax)
    dylm = (ylmp - ylmm) / delta
    return dylm


def get_ylm_ad(r, lmax):
    nlm = (lmax + 1) * (lmax + 1)
    rnorm = np.linalg.norm(r, axis=1)
    rhat = r / rnorm[:, np.newaxis]
    N = r.shape[0]
    print("N", N)
    ylm = np.zeros((N, nlm))
    dylm_tmp = np.zeros((N, 3, nlm))
    dylm = np.empty((4, N, nlm))
    libcider.recursive_sph_harm_deriv_vec(
        ctypes.c_int(nlm),
        ctypes.c_int(N),
        rhat.ctypes.data_as(ctypes.c_void_p),
        ylm.ctypes.data_as(ctypes.c_void_p),
        dylm_tmp.ctypes.data_as(ctypes.c_void_p),
    )
    dylm[0] = ylm
    dylm[1:] = dylm_tmp.transpose(1, 0, 2)
    subval = np.einsum("vgl,gv->gl", dylm[1:], rhat)
    print("SUBVAL", np.linalg.norm(subval))
    for l in range(lmax + 1):
        rlm1 = rnorm ** (l - 1)
        rl = rnorm**l
        for m in range(2 * l + 1):
            lm = l * l + m
            dylm[0, :, lm] *= rl
            dylm[1:, :, lm] *= rlm1
            dylm[1:, :, lm] += ylm[:, lm] * l * rlm1 * rhat.T
    xyz = dylm[..., 1:4].copy()
    dylm[..., 3] = xyz[..., 0]
    dylm[..., 1] = xyz[..., 1]
    dylm[..., 2] = xyz[..., 2]
    return dylm


def get_deriv_ylm_coeff(lmax):
    nlm = (lmax + 1) * (lmax + 1)
    gaunt_coeff = np.zeros((5, nlm))
    for l in range(lmax + 1):
        fe = clebsch_gordan_e3nn(l, 1, l + 1)
        for im in range(2 * l + 1):
            lm = l * l + im
            m = abs(im - l)
            l1m = l + 1 - m
            fac = ((2 * l + 3) * (l + 1)) ** 0.5
            gaunt_coeff[0, lm] = fac * fe[im, 2, im]
            gaunt_coeff[1, lm] = fac * fe[im, 2, im + 2]
            gaunt_coeff[2, lm] = fac * fe[im, 0, 2 * l - im]
            gaunt_coeff[3, lm] = fac * fe[im, 0, 2 * l - im + 2]
            gaunt_coeff[4, lm] = np.sqrt(
                ((2.0 * l + 3) * l1m * (l + 1 + m) / (2 * l + 1))
            )
    return gaunt_coeff


def get_deriv_ylm_coeff_v2(lmax):
    nlm = (lmax + 1) * (lmax + 1)
    gaunt_coeff = np.zeros((5, nlm))
    for l in range(lmax + 1):
        fe = clebsch_gordan_e3nn(l, 1, l + 1)
        for im in range(2 * l + 1):
            lm = l * l + im
            m = abs(im - l)
            l1m = l + 1 - m
            fac = ((2 * l + 3) * (l + 1)) ** 0.5
            gaunt_coeff[0, lm] = fac * fe[im, 2, im]
            gaunt_coeff[1, lm] = fac * fe[im, 2, im + 2]
            gaunt_coeff[2, lm] = fac * fe[im, 0, 2 * l - im]
            gaunt_coeff[3, lm] = fac * fe[im, 0, 2 * l - im + 2]
            # gaunt_coeff[:4, lm] *= np.sqrt(2 * m + 1) * (-1)**l
            gaunt_coeff[4, lm] = np.sqrt(
                ((2.0 * l + 3) * l1m * (l + 1 + m) / (2 * l + 1))
            )
    return gaunt_coeff


if __name__ == "__main__":
    from numpy.testing import assert_allclose, assert_almost_equal

    np.random.seed(32)
    lmax = 10
    N = 15
    nlm = (lmax + 1) * (lmax + 1)
    r = np.random.normal(size=(N, 3))
    dylm_ref = np.zeros((3, N, nlm))
    ylm0 = get_ylm(r, lmax)
    for i in range(3):
        dylm_ref[i] = get_ylm_fd(i, r, lmax)
    dylm_ref2 = get_ylm_ad(r, lmax)
    for lm in range(0, nlm):
        assert_allclose(ylm0[:, lm], dylm_ref2[0, :, lm], rtol=1e-6, atol=1e-5)
        assert_allclose(dylm_ref[0, :, lm], dylm_ref2[1, :, lm], rtol=1e-6, atol=1e-5)
        assert_allclose(dylm_ref[1, :, lm], dylm_ref2[2, :, lm], rtol=1e-6, atol=1e-5)
        assert_allclose(dylm_ref[2, :, lm], dylm_ref2[3, :, lm], rtol=1e-6, atol=1e-5)

    import time

    t0 = time.monotonic()
    gaunt_coeff = get_deriv_ylm_coeff(lmax)
    t1 = time.monotonic()
    print("TIME TO GEN COEFF", t1 - t0)
    print("GAUNT", gaunt_coeff)
    ylm0_fit = np.zeros((4, N, nlm))
    igrad = 0
    for l in range(lmax):
        mmax = 2 * l + 1
        for m in range(2 * l + 1):
            """
            if igrad == 0:
                lm = l * l + m
            else:
                lm = (l+1) * (l+1) - 1 - m
            lm0 = lm + 2 * l + 1
            lm1 = lm + 2 * l + 3
            ylm0_fit[0, :, lm0] = ylm0[:, lm]
            ylm0_fit[1, :, lm1] = ylm0[:, lm]
            """
            lm = l * l + m
            if igrad == 0:
                lmt = lm
            else:
                lmt = (l + 1) * (l + 1) - 1 - m
            lm0 = lmt + 2 * l + 1
            lm1 = lmt + 2 * l + 3
            # lm2 = lmh + 2 * l + 1
            # lm3 = lmh + 2 * l + 3
            if igrad == 0:
                ylm0_fit[0, :, lm0] = ylm0[:, lm]
                ylm0_fit[1, :, lm1] = ylm0[:, lm]
            else:
                ylm0_fit[0, :, lm0] = ylm0[:, lm]
                ylm0_fit[1, :, lm1] = ylm0[:, lm]
    for l in range(lmax + 1):
        mmax = 2 * l + 1
        for m in range(2 * l + 1):
            lm = l * l + m
            X = ylm0_fit[:, :, lm].T
            X = []
            inds = []
            mm = m - l
            am = abs(mm)
            f1 = (
                clebsch_gordan(l - 1, 1, l, mm - 1, 1, mm).n(64)
                * ((2 * l + 1) * l) ** 0.5
                / (2) ** 0.5
            )
            f2 = (
                clebsch_gordan(l - 1, 1, l, mm + 1, -1, mm).n(64)
                * ((2 * l + 1) * l) ** 0.5
                / (2) ** 0.5
                * -1
            )
            if l != 0:
                fe = clebsch_gordan_e3nn(l - 1, 1, l)
                print(mm, m, l, fe.shape)
                # print(fe)
                fac = ((2 * l + 1) * l) ** 0.5
                if igrad == 0:
                    igg = 2
                    mmm = m
                    m1m = m - 2
                    m1p = m
                elif igrad == 1:
                    igg = 0
                    mmm = m
                    # fac *= -1
                    m1p = 2 * l - m - 2
                    m1m = 2 * l - m
                else:
                    igg = 1
                    mmm = m
                    m1m = m - 1
                    m1p = m - 1
                # igg = 2 if igrad == 0 else 0
                # mmm = m if igrad == 0 else 2 * l - m
                fxp = fe[m1p, igg, mmm] if mm < l - 1 else 0
                fxp *= fac
                fxm = fe[m1m, igg, mmm] if mm > -l + 1 else 0
                fxm *= fac
            else:
                fxp = fxm = 0
            if mm == 0:
                f1 *= np.sqrt(2)
                f2 *= np.sqrt(2)
            # print('FACTORC', float(f1), float(f2))
            print("FACTORB", fxp, fxm)
            if igrad == 1:
                mvals = [
                    mm + 1,
                    mm - 1,
                ]
            else:
                mvals = [
                    -mm - 1,
                    -mm + 1,
                ]
            # print('MVALS', mvals)
            for i in range(2):
                # print('MM', m, mm, mvals[i])
                if abs(mvals[i]) >= l:
                    continue
                if np.linalg.norm(ylm0_fit[i, :, lm]) == 0:
                    continue
                for x in X:
                    if np.linalg.norm(x - ylm0_fit[i, :, lm]) < 1e-6:
                        break
                else:
                    inds.append((i, mvals[i]))
                    X.append(ylm0_fit[i, :, lm])
            if len(X) == 0:
                continue
            X = np.array(X).T
            y = dylm_ref[igrad, :, lm]
            # print(X.shape, y.shape)
            beta = np.linalg.inv(X.T.dot(X) + 1e-10 * np.identity(X.shape[1]))
            beta = beta.dot(X.T.dot(y))
            err = np.linalg.norm(X.dot(beta) - y)
            assert_almost_equal(err, 0, 4)
            print(l, m - l, inds, lm, np.round(beta, 6), err)
            # print(gaunt_coeff[2, lm], err)
    for lm in range(nlm):
        c_xl = np.zeros((N, nlm))
        c_xl[:, lm] = 1
        dylm_pred = np.zeros((3, N, nlm))
        libcider.fill_sph_harm_deriv_coeff(
            c_xl.ctypes.data_as(ctypes.c_void_p),
            dylm_pred.ctypes.data_as(ctypes.c_void_p),
            gaunt_coeff.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(N),
            ctypes.c_int(lmax),
        )
        dylm_pred = (dylm_pred * ylm0).sum(-1)
        print(lm)
        # print(r[:, 2])
        # print(ylm0[:, lm])
        # print(dylm_pred[2])
        # print(dylm_ref[2, :, lm])
        assert_almost_equal(dylm_pred[2], dylm_ref2[3, :, lm], 6)
        assert_almost_equal(dylm_pred[0], dylm_ref2[1, :, lm], 6)
        assert_almost_equal(dylm_pred[1], dylm_ref2[2, :, lm], 6)
